from __future__ import annotations

import tempfile
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from io import BufferedReader
from pathlib import Path, PurePosixPath
from typing import (IO, TYPE_CHECKING, AsyncContextManager, AsyncIterable, AsyncIterator, Awaitable, BinaryIO, Callable,
                    Mapping, NamedTuple, Sequence, Union, overload)

import pkg_resources
from typing_extensions import Literal, TypeAlias

from dagon import util

from .. import fs, task, ui
from ..event import Handler
from ..event.cancel import CancellationToken, raise_if_cancelled
from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..util import AsyncNullContext
from . import _cache

try:
    import aiohttp
except ModuleNotFoundError as e_:
    if e_.name != 'aiohttp':
        raise
    if not TYPE_CHECKING:
        aiohttp = None


class DownloadProgress(NamedTuple):
    n_bytes: int
    total_bytes: int | None
    progress: float | None


def is_supported() -> bool:
    try:
        pkg_resources.require('dagon[http]')
        return True
    except pkg_resources.DistributionNotFound:
        return False


def check_supported() -> None:
    try:
        ws = pkg_resources.WorkingSet()
        ws.require('dagon[http]')
    except pkg_resources.DistributionNotFound as e:
        raise RuntimeError('Install Dagon with the `http` extra to use the '
                           'HTTP functionality (Install as `dagon[http]`)') from e


@dataclass
class _GlobalContext:
    session: aiohttp.ClientSession
    cache: _cache.CacheAccess


@asynccontextmanager
async def _new_context() -> AsyncIterator[_GlobalContext]:
    async with aiohttp.ClientSession() as s:
        yield _GlobalContext(s, _cache.open_user_cache())


@asynccontextmanager
async def _get_context() -> AsyncIterator[_GlobalContext]:
    check_supported()
    try:
        glb = _Ext.global_data()
    except LookupError:
        async with _new_context() as c:
            yield c
    else:
        assert glb
        yield glb


class _Ext(BaseExtension[None, '_GlobalContext | None', None]):
    dagon_ext_name: str = 'dagon.http'

    def global_context(self, graph: OpaqueTaskGraphView) -> AsyncContextManager[_GlobalContext | None]:
        if aiohttp is None:
            return AsyncNullContext(None)
        return _new_context()


_LowRequestBodyItem = Union[bytes, AsyncIterable[bytes], BinaryIO, None]


@asynccontextmanager
async def _init_request(
    method: str,
    url: str,
    headers: Mapping[str, str] | None,
    data: _LowRequestBodyItem,
) -> AsyncIterator[aiohttp.ClientResponse]:
    check_supported()
    headers = headers or {}
    async with _get_context() as glb:
        async with glb.session.request(method, url, data=data, headers=headers) as res:
            res.raise_for_status()
            yield res


class _CacheLookupResult(NamedTuple):
    use: _cache.CacheEntry | None
    revalidate: Sequence[_cache.CacheEntry]


class _RequestContext(NamedTuple):
    cache: _cache.CacheAccess
    url: str
    headers: Mapping[str, str]
    cancel: CancellationToken | None
    on_progress: Handler[DownloadProgress] | None


def _try_find_cached(cache: _cache.CacheAccess, url: str) -> _CacheLookupResult:
    all_for_url = cache.for_url(url)
    revalidate: list[_cache.CacheEntry] = []
    for c in all_for_url:
        if c.cache_control.must_revalidate or c.is_stale:
            revalidate.append(c)
        elif c.is_fresh or c.cache_control.immutable:
            # Cache entry is fresh and lacks must-revalidate, so it can always be used
            return _CacheLookupResult(c, [])
        elif c.etag or c.last_modified:
            # Cache entry is not fresh and not stale, so it lacks a max-age. We can still validate
            # using Etag or modification time, though
            revalidate.append(c)
        else:
            # No max-age, no Etag, and no Last-Modified. This cache entry is useless.
            pass
    return _CacheLookupResult(None, revalidate)


async def _get_chunked_with_revalidate(req: _RequestContext,
                                       cached_revalidate: Sequence[_cache.CacheEntry]) -> AsyncIterator[bytes]:
    headers = dict(req.headers)
    if len(cached_revalidate) == 1 and cached_revalidate[0].last_modified:
        headers['If-Modified-Since'] = cached_revalidate[0].last_modified
    else:
        headers['If-None-Match'] = ', '.join(c.etag for c in cached_revalidate if c.etag)
    async with AsyncExitStack() as stack:
        resp = await stack.enter_async_context(_init_request('GET', req.url, headers, None))
        if resp.status == 304:
            resp_etag = resp.headers.get('etag', None)
            # Cache hit!
            chosen: _cache.CacheEntry | None = None
            if resp_etag:
                chosen = util.first((c for c in cached_revalidate if c.etag == resp_etag), default=None)
            if not chosen:
                chosen = util.first(cached_revalidate, default=None)
            if chosen:
                return await _get_chunked_yield_cached(chosen, req)
            # Didn't find anything?
            raise RuntimeError(f'Server responded with "304 Not Modified", but we do '
                               f'not have a cached response [url={req.url}]')
        else:
            return _yield_resp(resp, req, stack.pop_all())


async def _yield_resp(resp: aiohttp.ClientResponse, req: _RequestContext,
                      stack: AsyncExitStack) -> AsyncIterator[bytes]:
    async with stack:
        raise_if_cancelled(req.cancel)
        dltime = datetime.now()
        cl = resp.headers.get('content-length', None)
        total_bytes: int | None
        try:
            total_bytes = int(cl) if cl else None
        except ValueError:
            total_bytes = None
        n_bytes = 0
        do_cache = True
        etag = resp.headers.get('Etag')
        cc = resp.headers.get('Cache-Control')
        lmod = resp.headers.get('Last-Modified')
        try:
            age = int(resp.headers.get('Age', '0'))
        except ValueError:
            age = 0
        if (etag, cc, lmod) == (None, None, None):
            do_cache = False
        tf: IO[bytes] | None = None
        tname = None
        if do_cache:
            req.cache.root.mkdir(exist_ok=True, parents=True)
            tfd, tname = tempfile.mkstemp('.dl', dir=req.cache.root, text=False)
            tf = stack.enter_context(open(tfd, 'wb'))
            stack.push_async_callback(lambda: fs.remove(tname, absent_ok=True))
        async for dat in resp.content.iter_chunked(1024 * 1024):
            raise_if_cancelled(req.cancel)
            n_bytes += len(dat)
            if req.on_progress:
                progress = None if total_bytes is None else (n_bytes / total_bytes)
                req.on_progress(DownloadProgress(n_bytes, total_bytes, progress))
            if tf:
                tf.write(dat)
            yield dat
        if tf:
            assert do_cache
            tf.close()
            assert tname
            await req.cache.save(url=req.url,
                                 last_modified=resp.headers.get('Last-Modified'),
                                 etag=resp.headers.get('Etag'),
                                 age=age,
                                 cache_control=resp.headers.get('Cache-Control'),
                                 download_time=dltime,
                                 take_file=Path(tname))


async def _get_chunked_try_cached(req: _RequestContext) -> AsyncIterator[bytes]:
    cached = _try_find_cached(req.cache, req.url)
    if cached.use:
        return await _get_chunked_yield_cached(cached.use, req)
    return await _get_chunked_with_revalidate(req, cached.revalidate)


async def _get_chunked_yield_cached(cached: _cache.CacheEntry, req: _RequestContext) -> AsyncIterator[bytes]:
    fpath = req.cache.path_to(cached)
    try:
        fd = fpath.open('rb')
        return _agen_read(fd)
    except FileNotFoundError:
        await req.cache.drop(cached)
        return await _get_chunked_try_cached(req)


async def _agen_read(fd: BufferedReader) -> AsyncIterator[memoryview]:
    with fd:
        async for dat in fs.read_blocks_from(fd):
            yield dat


async def get_chunked(url: str,
                      *,
                      headers: Mapping[str, str] | None = None,
                      cancel: CancellationToken | None = None,
                      on_progress: Handler[DownloadProgress] | None = None) -> AsyncIterator[bytes | memoryview]:
    """
    Asynchronously download byte chunks from the remote URL.

    :param url: The URL to download from
    :param cancel: A cancellation token to cancel the operation
    :param on_progress: Optional event to dispatch progress

    Use this function as an asynchronous iterator::

        async for buf in dagon.http.get_chunked('http://example.com/big-file'):
            process_bytes(buf)
    """
    cancel = cancel or CancellationToken.get_context_local()
    gen: AsyncIterator[bytes | memoryview]
    async with _get_context() as ctx:
        req = _RequestContext(ctx.cache, url, headers or {}, cancel, on_progress)
        gen = await _get_chunked_try_cached(req)
        async for part in gen:
            yield part


@asynccontextmanager
async def download_tmp(url: str,
                       *,
                       cancel: CancellationToken | None = None,
                       on_progress: Handler[DownloadProgress] | None = None,
                       delete: bool = True,
                       tmp_dir: fs.Pathish | None = None,
                       suffix: str | None = None) -> AsyncIterator[Path]:
    """
    Download the given URL and store the content in a temporary file, yielding
    the path and (optionally) deleting it when done.

    :param url: The URL to download from
    :param cancel: A cancellation token to cancel the operation
    :param progress: Optional event to dispatch progress
    :param delete: If ``True``, the file will be deleted when the context
        manager exits.
    :param tmp_dir: Override the directory that will contain the temporary file.
        Will use the system's default temporary file directory otherwise.
    :param suffix: Set the filename suffix for the generated temporary file.

    :rtype: AsyncContextManager[Path]

    Use this as an asynchronous context manager::

        async with dagon.http.download_tmp(big_file_url) as local_fpath:
            process_file(local_fpath)
    """

    with tempfile.NamedTemporaryFile('wb', delete=False, dir=None if tmp_dir is None else str(tmp_dir),
                                     suffix=suffix) as tmpfd:
        try:
            async for dat in get_chunked(url, cancel=cancel, on_progress=on_progress):
                tmpfd.write(dat)
        except:
            tmpfd.close()
            if not delete:
                Path(tmpfd.name).unlink()
            raise
        else:
            tmpfd.close()
            yield Path(tmpfd.name)
        finally:
            if delete and Path(tmpfd.name).exists():
                Path(tmpfd.name).unlink()


async def download(url: str,
                   *,
                   destination: fs.Pathish,
                   if_exists: fs.IfExists = 'fail',
                   on_progress: Handler[DownloadProgress] | None = None,
                   cancel: CancellationToken | None = None,
                   use_parent_as_tmp_dir: bool = False) -> Path:
    """
    Download the given URL to the given file destination.

    :param url: The URL to download
    :param destination: The file path where to store the downloaded file
    :param if_exists: The policy of behavior when the file already exists
    :param cancel: A cancellation token to cancel the operation
    :param progress: Optional event to dispatch progress
    :param use_parent_as_tmp_dir: If ``True``, the destination directory will
        be used as the temporary directory during the download.

    .. note::
        If ``if_exists`` is set to `"keep"` and the destination
        already exists, this function will return immediately without
        performing any HTTP operations. This makes it useful to perform "lazy"
        downloads without additional branching and logic.
    """
    destination = Path(destination)
    if destination.exists():
        if if_exists == 'fail':
            raise RuntimeError(f'Download destination {destination} already exists')
        if if_exists == 'keep':
            return destination
        if if_exists == 'replace':
            await fs.remove(destination, cancel=cancel, recurse=True)
    dest = Path(destination)
    tmp_dir = None
    if use_parent_as_tmp_dir:
        tmp_dir = str(dest.parent)

    async with download_tmp(
            url,
            on_progress=on_progress,
            cancel=cancel,
            delete=False,
            tmp_dir=tmp_dir,
    ) as tmp_path:
        dest.parent.mkdir(parents=True, exist_ok=True)
        await fs.safe_move_file(tmp_path, dest, mkdirs=True, cancel=cancel)
    return dest


_StrOrStrFactory: TypeAlias = 'str | Callable[[], str | Awaitable[str]]'


@overload
def download_task(name: str,
                  url: _StrOrStrFactory,
                  *,
                  doc: str = ...,
                  default: bool = ...,
                  message: str = ...,
                  depends: task.DependsArg = ...,
                  order_only_depends: task.DependsArg = ...,
                  filename: str = ...) -> task.Task[Path]:
    ...


@overload
def download_task(name: str,
                  url: _StrOrStrFactory,
                  *,
                  doc: str = ...,
                  default: bool = ...,
                  message: str = ...,
                  depends: task.DependsArg = ...,
                  order_only_depends: task.DependsArg = ...,
                  destination: fs.Pathish,
                  if_exists: Literal['keep', 'replace', 'fail'] = 'keep') -> task.Task[Path]:
    ...


def download_task(name: str,
                  url: _StrOrStrFactory,
                  *,
                  doc: str | None = None,
                  default: bool = False,
                  message: str | None = None,
                  depends: task.DependsArg = (),
                  order_only_depends: task.DependsArg = (),
                  destination: fs.Pathish | None = None,
                  filename: str | None = None,
                  if_exists: Literal['keep', 'replace', 'fail'] = 'keep') -> task.Task[Path]:
    """
    Define a task that downloads a resource from the given URL. The result of
    the task is the path to a local copy of the downloaded resource.

    :param name: The name of the task to generate.
    :param url: A URL string, or a factory function that returns a URL string.
        The resource at this URL will be downloaded.
    :param message: A status message to display while downloading the file.
    :param destination: The path to where the downloaded file will be written.
        If omitted, the file will be downloaded to a temporary location and
        removed after the graph execution completes.
    :param filename: Set the filename of the downloaded file. By default, generates
        a filename based on the URL string.
    :param if_exists: If ``destination`` was provided and the destination names
        an existing path, what should be done? Default is ``"keep"`` which will
        leave the destination unmodified and not download any data.
    """
    url_ = url
    dest_ = destination

    @task.define(
        name=name,
        default=default,
        depends=depends,
        order_only_depends=order_only_depends,
        doc=doc,
    )
    async def fn() -> Path:
        url = url_
        destination = dest_
        if not isinstance(url, str):
            url = url()
            if not isinstance(url, str):
                url = await url
        _: str = url
        if destination is None:
            tdir = tempfile.mkdtemp()
            destination = Path(tdir) / (filename or PurePosixPath(url).name or 'download')
        destination = Path(destination)
        if destination.is_file():
            if if_exists == 'keep':
                return destination
            if if_exists == 'fail':
                raise FileExistsError(destination)
            _replace: Literal['replace'] = if_exists
            await fs.remove(destination)
        ui.status(message or f'Downloading [{url}]')
        await download(url, destination=destination, on_progress=lambda pr: ui.progress(pr.progress))
        ui.progress(None)
        return destination

    return fn



from __future__ import annotations

import hashlib
from datetime import datetime, timedelta
from pathlib import Path, PurePosixPath
from typing import Iterable, NamedTuple, cast
from urllib.parse import quote as url_quote
from urllib.parse import urlparse

from .. import db as db_mod
from .. import fs
from ..util.paths import user_cache_path


class CacheAccess:
    def __init__(self, files_root: Path, db: db_mod.Database) -> None:
        self._files_root = files_root
        self._db = db

    @property
    def root(self) -> Path:
        """The root directory of all cached data"""
        return self._files_root

    def for_url(self, url: str) -> Iterable[CacheEntry]:
        q = cast(
            'Iterable[tuple[int, str, str | None, str | None, str | None, int | None, int]]',
            self._db(
                r'''
            SELECT resource_id,
                   url,
                   etag,
                   cache_control,
                   last_modified,
                   age,
                   download_time
              FROM dagon_http_cache
             WHERE url = :url
            ''',
                url=url,
            ))
        for res_id, u, etag, cache_control, last_mod, age, dl_time in q:
            yield CacheEntry(res_id, u, etag,
                             CacheControl.parse(cache_control) if cache_control else NIL_CACHE_CONTROL, last_mod, age,
                             datetime.fromtimestamp(dl_time))

    async def drop(self, entry: CacheEntry) -> None:
        async with self._db.transaction_context():
            fpath = self.path_to(entry)
            await fs.remove(fpath, absent_ok=True)
            self._db(r'DELETE FROM dagon_http_cache WERE resource_id=:id', id=entry.resource_id)

    async def save(self, *, url: str, last_modified: str | None, etag: str | None, cache_control: str | None,
                   age: int | None, download_time: datetime, take_file: Path) -> None:
        if (last_modified, etag, cache_control) == (None, None, None):
            raise TypeError('Insufficient arguments to perform caching (Need at least '
                            'one of last_modified, etag, or cache_control)')
        async with self._db.transaction_context():
            c = self._db(
                r'''
                INSERT INTO dagon_http_cache
                    (url, etag, cache_control, last_modified, age, download_time)
                VALUES
                    (:url, :etag, :cc, :lmod, :age, :dtime)
                ''',
                url=url,
                etag=etag,
                cc=cache_control,
                lmod=last_modified,
                age=age,
                dtime=download_time.timestamp(),
            )
            rowid = c.lastrowid
            assert rowid
            cachepath = cachepath_for_url(url, rowid)
            dest = self._files_root / cachepath
            await fs.safe_move_file(take_file, dest, if_exists='replace', mkdirs=True)

    def path_to(self, entry: CacheEntry) -> Path:
        return self._files_root / cachepath_for_url(entry.url, entry.resource_id)


def cachepath_for_url(url: str, rowid: int) -> str:
    urltup = urlparse(url)
    safepath = mangle_path(urltup.path)
    x = hashlib.md5(url.encode()).hexdigest()[:4]
    return str(PurePosixPath(f'{urltup.scheme}/{urltup.hostname}-{urltup.port or "nil"}/{x}/{safepath}-{rowid}'))


def mangle_path(p: str) -> PurePosixPath:
    parts = PurePosixPath(p).parts
    if parts and parts[0] == '/':
        parts = parts[1:]
    if len(parts) == 0:
        return PurePosixPath('_')
    par = PurePosixPath(*(mangle_path_part(p) for p in parts[:-1]))
    return par / mangle_path_part(parts[-1])


def mangle_path_part(p: str) -> str:
    q = url_quote(p)
    if len(q) > 8:
        q = q[:8]
    return q


class CacheEntry(NamedTuple):
    resource_id: int
    url: str
    etag: str | None
    cache_control: CacheControl
    last_modified: str | None
    age_offset: int | None
    download_time: datetime

    @property
    def age(self) -> timedelta:
        since_req = datetime.now() - self.download_time
        return since_req + timedelta(seconds=self.age_offset or 0)

    @property
    def has_max_age(self) -> bool:
        return self.cache_control.max_age is not None

    @property
    def is_fresh(self) -> bool:
        """Whether the cache entry is "fresh" according to its Cache-Control headers"""
        return (self.cache_control.max_age is not None) and (self.age.total_seconds() < self.cache_control.max_age)

    @property
    def is_stale(self) -> bool:
        """Whether the entry is expired at the current time"""
        return (self.cache_control.max_age is not None) and (self.age.total_seconds() >= self.cache_control.max_age)

    @property
    def expires_at(self) -> datetime | None:
        """The datetime at which this cache entry will be considered expired"""
        if self.cache_control.max_age is None:
            return None
        return self.download_time + timedelta(seconds=self.cache_control.max_age) - timedelta(
            seconds=self.age_offset or 0)


def open_user_cache() -> CacheAccess:
    http_dir: Path = user_cache_path() / 'dagon/http'
    return open_cache_in(http_dir)


def open_cache_in(dirpath: Path) -> CacheAccess:
    dirpath.mkdir(exist_ok=True, parents=True)
    db = db_mod.Database.open(dirpath / 'data.db')
    db_mod.apply_migrations(db.sqlite3_db, 'dagon_http_meta', [
        r'''
        CREATE TABLE dagon_http_cache (
            resource_id INTEGER PRIMARY KEY,
            url TEXT NOT NULL,
            etag TEXT,
            cache_control TEXT,
            last_modified TEXT,
            age INTEGER,
            download_time INTEGER NOT NULL,
            CONSTRAINT has_cache_attr CHECK (
               (etag NOT NULL) +
               (cache_control NOT NULL) +
               (last_modified NOT NULL) > 0
            )
        )
        ''',
        r'CREATE INDEX idx_dagon_http_cache_by_etag ON dagon_http_cache (etag)',
        r'CREATE INDEX idx_dagon_http_cache_by_url ON dagon_http_cache(url)',
    ])
    return CacheAccess(dirpath / 'files', db)


class CacheControl(NamedTuple):
    private: bool
    public: bool
    no_store: bool
    no_cache: bool
    immutable: bool
    max_age: int | None
    must_revalidate: bool

    @staticmethod
    def parse(hdr: str) -> CacheControl:
        items = hdr.split(',')
        items = [s.strip().lower() for s in items]
        max_age: int | None = None
        prefix = 'max-age='
        for i in items:
            if not i.startswith(prefix):
                continue
            tail = i[len(prefix):]
            try:
                max_age = int(tail)
            except ValueError:
                pass
            break
        return CacheControl(private='private' in items,
                            public='public' in items,
                            no_store='no-store' in items,
                            no_cache='no-cache' in items,
                            max_age=max_age,
                            immutable='immutable' in items,
                            must_revalidate='must-revalidate' in items)


NIL_CACHE_CONTROL = CacheControl(False, False, False, False, False, None, False)



"""
Module for the Dagon persistent database
"""

from __future__ import annotations

import argparse
import asyncio
import dataclasses
import datetime
import enum
import hashlib
import json
import os
import random
import sqlite3
from contextlib import ExitStack, asynccontextmanager, nullcontext
from pathlib import Path, PurePath, PurePosixPath
from typing import (Any, AsyncGenerator, AsyncIterator, ContextManager, Iterable, NamedTuple, NewType, Optional,
                    Sequence, Union, cast)

from typing_extensions import Literal, Protocol, TypeAlias

from .. import util
from ..core.result import Cancellation, Failure, NodeResult, Success
from ..event import Event, events
from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..fs import Pathish
from ..proc import CompletedProcess
from ..task.dag import OpaqueTask
from ..util import Opaque, fixup_dataclass_docs

QueryParameter = Union[int, str, bytes, float, None]
"""Any type that is known-safe to use as a database query parameter"""


class TaskState(enum.Enum):
    """The state of a task"""
    Pending = 'pending'
    'Task is enqueued, but has not started'
    Running = 'running'
    'Task is running'
    Succeeded = 'succeeded'
    'Task executed to completion without error'
    Failed = 'failed'
    'Task raised an exception'
    Cancelled = 'cancelled'
    'Task was cancelled'


class DatabaseError(RuntimeError):
    """
    Exception thrown for database errors
    """


class MissingDatabase(DatabaseError):
    """
    Exception thrown when the requested database file is absent.
    """
    def __init__(self, path: Path) -> None:
        super().__init__(f'Missing database file at "{path}"')
        self.expected_path = path


class InvalidDatabase(DatabaseError):
    """
    Exception thrown if an existing database cannot be ready by Dagon
    """
    def __init__(self, path: Path) -> None:
        super().__init__(f'Not a Dagon database file at "{path}"')
        self.path = path


class _ProcOutputItem(Protocol):
    out: bytes
    kind: Literal['error', 'output']


def _exec_kw(db: Union[sqlite3.Connection, sqlite3.Cursor], stmt: str, **kwargs: QueryParameter) -> sqlite3.Cursor:
    return db.execute(stmt, kwargs)


def _db_connect(db_path: Pathish) -> sqlite3.Connection:
    """Open a SQLite database connection. Sets appropriate database options."""
    db = sqlite3.connect(str(db_path), isolation_level=None)
    db.row_factory = sqlite3.Row
    db.executescript(r'''
        PRAGMA foreign_keys = ON;
        PRAGMA temp_store = MEMORY;
        PRAGMA journal_mode = MEMORY;
        PRAGMA synchronous = OFF;
    ''')
    return db


def apply_migrations(db: sqlite3.Connection, meta_table: str, migrations: Sequence[str]) -> None:
    meta_init = fr'''
        CREATE TABLE IF NOT EXISTS {meta_table} (
            version INTEGER NOT NULL
        )
    '''
    db.execute(meta_init)
    meta_rows = list(db.execute(f'SELECT version FROM {meta_table}'))
    version: int
    if not meta_rows:
        version = 0
    else:
        version = util.cell(meta_rows)
    n_migrations = len(migrations)
    if version == n_migrations:
        # All migrations are up-to-date
        return
    savepoint = f'save-{random.randint(0, 99999)}'
    db.execute(f"savepoint '{savepoint}'")
    try:
        _apply_migrations(db, version, migrations)
    except:
        db.execute(f"rollback to '{savepoint}'")
        db.execute(f"release savepoint '{savepoint}'")
        raise
    else:
        db.execute(f'INSERT OR REPLACE INTO {meta_table} (rowid, version) VALUES (1, ?)', [n_migrations])
        db.execute(f"release savepoint '{savepoint}'")


def _apply_migrations(db: sqlite3.Connection, cur_version: int, migrations: Sequence[str]) -> None:
    for m in migrations[cur_version:]:
        db.execute(m)


_DB_SCHEMA = r'''
CREATE TABLE dagon_meta (
    hash TEXT NOT NULL
);
CREATE TABLE dagon_runs (
    run_id INTEGER PRIMARY KEY,
    time REAL NOT NULL
);
CREATE TABLE dagon_tasks (
    task_id INTEGER PRIMARY KEY,
    name TEXT NOT NULL UNIQUE
);
CREATE TABLE dagon_task_runs (
    task_run_id INTEGER PRIMARY KEY,
    task_id
        NOT NULL REFERENCES dagon_tasks
        ON DELETE CASCADE,
    run_id
        NOT NULL REFERENCES dagon_runs
        ON DELETE CASCADE,
    end_state TEXT DEFAULT 'pending',
    start_time REAL NOT NULL,
    duration REAL,
    UNIQUE(task_id, run_id) ON CONFLICT ABORT
);
CREATE TABLE dagon_task_events (
    task_event_id INTEGER PRIMARY KEY,
    task_run_id
        NOT NULL REFERENCES dagon_task_runs
        ON DELETE CASCADE,
    event NOT NULL,
    time REAL NOT NULL
);
CREATE TABLE dagon_proc_execs (
    proc_exec_id INTEGER PRIMARY KEY,
    task_run_id
        NOT NULL REFERENCES dagon_task_runs
        ON DELETE CASCADE,
    cmd NOT NULL,
    start_cwd TEXT NOT NULL,
    stdout BLOB,
    stderr BLOB,
    retc NOT NULL,
    start_time REAL NOT NULL,
    duration REAL NOT NULL
);
CREATE TABLE dagon_task_deps_rel (
    tree_id INTEGER PRIMARY KEY,
    run_id
        NOT NULL REFERENCES dagon_runs
        ON DELETE CASCADE,
    dependent TEXT NOT NULL,
    depends_on TEXT,
    UNIQUE(run_id, dependent, depends_on) ON CONFLICT REPLACE
);
CREATE TABLE dagon_intervals (
    interval_id INTEGER PRIMARY KEY,
    task_run_id REFERENCES dagon_task_runs ON DELETE CASCADE,
    label TEXT NOT NULL,
    meta,
    start_time REAL NOT NULL,
    end_time REAL
);
'''


def schema_hash() -> str:
    """
    Hash of the database schema. Used to version the database until we someday
    need to do real database migrations.
    """
    return hashlib.sha256(_DB_SCHEMA.encode('utf-8')).hexdigest()


def _create_sqlite_db(db_path: Pathish) -> sqlite3.Connection:
    db = _db_connect(db_path)
    apply_migrations(db, 'dagon_db_meta', [
        r'CREATE TABLE dagon_runs (run_id INTEGER PRIMARY KEY, time REAL NOT NULL)',
        r'CREATE TABLE dagon_tasks (task_id INTEGER PRIMARY KEY, name TEXT NOT NULL UNIQUE)',
        r'''
        CREATE TABLE dagon_task_runs (
            task_run_id INTEGER PRIMARY KEY,
            task_id INTEGER NOT NULL
                    REFERENCES dagon_tasks ON DELETE CASCADE,
            run_id INTEGER NOT NULL
                   REFERENCES dagon_runs ON DELETE CASCADE,
            end_state TEXT DEFAULT 'pending',
            start_time REAL NOT NULL,
            duration REAL,
            UNIQUE (task_id, run_id) ON CONFLICT ABORT
        )
        ''',
        r'''
        CREATE TABLE dagon_task_deps_rel (
            tree_id INTEGER PRIMARY KEY,
            run_id INTEGER NOT NULL
                   REFERENCES dagon_runs ON DELETE CASCADE,
            dependent TEXT NOT NULL,
            depends_on TEXT,
            UNIQUE(run_id, dependent, depends_on) ON CONFLICT REPLACE
        )
        ''',
        r'''
        CREATE TABLE dagon_intervals (
            interval_id INTEGER PRIMARY KEY,
            task_run_id REFERENCES dagon_task_runs ON DELETE CASCADE,
            label TEXT NOT NULL,
            meta,
            start_time REAL NOT NULL,
            end_time REAL
        )
        ''',
        r'''
        CREATE TABLE dagon_proc_execs (
            proc_exec_id INTEGER PRIMARY KEY,
            task_run_id
                NOT NULL REFERENCES dagon_task_runs
                ON DELETE CASCADE,
            cmd NOT NULL,
            start_cwd TEXT NOT NULL,
            stdout BLOB,
            stderr BLOB,
            retc NOT NULL,
            start_time REAL NOT NULL,
            duration REAL NOT NULL
        )''',
    ])
    # db.executescript(_DB_SCHEMA)
    # _exec_kw(db, 'INSERT INTO dagon_meta VALUES (:hash)', hash=schema_hash())
    return db


def open_sqlite_db(db_path: Pathish,
                   *,
                   mode: Literal['create-if-missing', 'force-create-new',
                                 'open-existing'] = 'create-if-missing') -> sqlite3.Connection:
    """
    Get an sqlite3 connection object that is set up and ready to use to persist
    information.

    If the given DB exists and has a matching schema version, it is used as-is.
    If the DB does not exist, or the DB has the wrong schema version, a new
    database is created at ``db_path``.
    """
    if db_path in ('', ':memory:'):
        db = _create_sqlite_db(db_path)
    elif mode == 'create-if-missing':
        db = _create_sqlite_db(db_path)
    elif mode == 'force-create-new':
        Path(db_path).unlink(missing_ok=True)
        db = _create_sqlite_db(db_path)
    elif mode == 'open-existing':
        if not Path(db_path).is_file():
            raise FileNotFoundError(db_path)
        db = _create_sqlite_db(db_path)
    else:
        _: Opaque = mode
        assert False
    return db


RunID = NewType('RunID', int)
TaskID = NewType('TaskID', int)
TaskRunID = NewType('TaskRunID', int)
FileID = NewType('FileID', int)
IntervalID = NewType('IntervalID', int)
ProcExecID = NewType('ProcExecID', int)

TaskRunRow: TypeAlias = 'tuple[TaskRunID, TaskID, RunID, int, str, float, Optional[float]]'


@dataclasses.dataclass(frozen=True)
class FileInfo:
    """
    Information about a file stored in the database
    """
    #: The Run ID that owns the file
    run_id: RunID
    #: The ID of the file object
    file_id: FileID
    #: The path to the file
    path: PurePosixPath
    #: The size of the file (in bytes). See :func:`pretty_size`
    size: int

    @property
    def pretty_size(self) -> str:
        """
        Return a prettified string representing the size of the file.
        """
        s = self.size
        for x in ('bytes', 'KiB', 'MiB', 'GiB', 'TiB'):
            if s < 1024:
                return f'{s:3.1f} {x}'
            s //= 1024
        raise RuntimeError('Value is too large. Wow.')


class Database:
    """
    Database access for persisting and obtaining data about executions.

    :param db: The SQLite database connection to use. Should have the schema
        already present.
    """
    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db
        self._transaction_lock = asyncio.Lock()

    @staticmethod
    def open(
            db_path: Pathish,
            *,
            mode: Literal['create-if-missing', 'force-create-new',
                          'open-existing'] = 'create-if-missing') -> 'Database':
        """
        Get or create a new database object at the given path (or ``:memory:``)
        """
        db = open_sqlite_db(db_path, mode=mode)
        return Database(db)

    @property
    def sqlite3_db(self) -> sqlite3.Connection:
        return self._db

    def new_run_id(self) -> RunID:
        """
        Create a new Run ID
        """
        c = self._db.cursor()
        _exec_kw(c, 'INSERT INTO dagon_runs (time) VALUES (:time)', time=datetime.datetime.now().timestamp())
        rowid = c.lastrowid
        assert rowid is not None
        cur = RunID(rowid)
        return cur

    def __call__(self, q: str, **kwargs: QueryParameter) -> sqlite3.Cursor:
        """
        Execute a database query ``q`` with the given keyword argument bindings.
        """
        return _exec_kw(self._db, q, **kwargs)

    @asynccontextmanager
    async def transaction_context(self) -> AsyncGenerator[None, None]:
        async with self._transaction_lock:
            with util.recursive_transaction(self.sqlite3_db):
                yield

    def get_task_rowid(self, task: str) -> TaskID:
        """
        Get the ID of the named task (inserting a new row if it does not
        already exist).
        """
        for tid, in self('SELECT task_id FROM dagon_tasks WHERE name = :name', name=task):
            return TaskID(tid)
        c = self._db.cursor()
        _exec_kw(c, 'INSERT INTO dagon_tasks (name) VALUES (:name)', name=task)
        rowid = c.lastrowid
        assert rowid is not None
        return TaskID(rowid)

    def add_task_run(self,
                     *,
                     run: RunID,
                     task: TaskID,
                     state: TaskState = TaskState.Pending,
                     start_time: Optional[datetime.datetime] = None) -> TaskRunID:
        """
        Create a new task run in the database.

        :param run: The run that owns the task run.
        :param task: The ID of the task (taken from :func:`get_task_rowid`)
        :param state: The initial state of the task
        :param start_time: The time at which the task started executing

        :returns: The new DB ID of the task run.
        """
        c = self._db.cursor()
        _exec_kw(
            c,
            'INSERT INTO dagon_task_runs (task_id, run_id, start_time, end_state) '
            'VALUES (:task_id, :run_id, :start_time, :state)',
            task_id=task,
            run_id=run,
            start_time=(start_time or datetime.datetime.now()).timestamp(),
            state=state.value,
        )
        trun_id = c.lastrowid
        assert trun_id is not None
        return TaskRunID(trun_id)

    def new_interval(self, trun_id: TaskRunID, label: str, time: Optional[datetime.datetime] = None) -> IntervalID:
        """
        Create a new interval.

        :param trun_id: The task run ID, taken from :func:`add_task_run`
        :param label: The label of the interval
        :param time: The start time of the interval. Default is the current time.

        :return: The DB ID of the interval.

        .. note:: Use :func:`set_interval_end` to mark the end time of the
            interval
        """
        time = time or datetime.datetime.now()
        c = self._db.cursor()
        _exec_kw(
            c,
            r'''
            INSERT INTO dagon_intervals (task_run_id, start_time, label, meta)
            VALUES (:trun_id, :time, :label, :meta)
            ''',
            trun_id=trun_id,
            time=time.timestamp(),
            label=label,
            meta='{}',
        )
        rowid = c.lastrowid
        assert rowid is not None
        return IntervalID(rowid)

    def set_interval_meta(self, interval: IntervalID, meta: Any) -> None:
        """
        Set a metadata value for the interval.

        :param interval: The DB ID obtained from :func:`new_interval`.
        :param meta: Any SQLite-serializable value to attach to the interval.
        """
        self(
            r'''
            UPDATE dagon_intervals
            SET meta = :meta
            WHERE interval_id = :id
            ''',
            id=interval,
            meta=json.dumps(meta),
        )

    def set_interval_end(self, interval: IntervalID, time: Optional[datetime.datetime] = None) -> None:
        """
        Mark the end time of an interval.

        :param interval: The DB ID obtained from :func:`new_interval`
        :param time: The end time of the interval. Defaults to the current time.
        """
        time = time or datetime.datetime.now()
        self(
            r'''
            UPDATE dagon_intervals
            SET end_time = :time
            WHERE interval_id = :id
            ''',
            time=time.timestamp(),
            id=interval,
        )

    def update_task_state(self,
                          trun_id: TaskRunID,
                          state: TaskState | None = None,
                          duration: datetime.timedelta | None = None) -> None:
        """
        Update the state of a task that has run.

        :param trun_id: The task run ID, obtained from :func:`add_task_run`.
        :param state: The new state. If ``None``, the state will not be changed.
        :param duration: The duration of execution. If ``None``, the duration
            will not be changed.
        """
        if state:
            self('UPDATE dagon_task_runs SET end_state = :state WHERE task_run_id = :id', state=state.value, id=trun_id)
        if duration is not None:
            self('UPDATE dagon_task_runs SET duration = :dur WHERE task_run_id = :id',
                 id=trun_id,
                 dur=duration.total_seconds())

    def store_proc_execution(self, trun_id: TaskRunID, *, cmd: Sequence[str], cwd: PurePath,
                             output: Iterable[_ProcOutputItem] | None, retc: int, start_time: datetime.datetime,
                             duration: float) -> ProcExecID:
        """
        Record the execution of a subprocess.

        :param trun_id: The task run ID, obtained from :func:`add_task_run`
        :param cmd: List of strings as the command line.
        :param cwd: The path to the starting working directory of the process.
        :param stdout: The stdout output from the process.
        :param stderr: The stderr output from the process.
        :param retc: The exit code of the process.
        :param state_time: The time that the process started.
        :param duration: The duration (in seconds) of the subprocess execution.

        :return: The DB ID referencing the process execution.
        """
        q = r'''
            INSERT INTO dagon_proc_execs (
                task_run_id,
                cmd,
                start_cwd,
                stdout,
                stderr,
                retc,
                start_time,
                duration
            ) VALUES (
                :trun_id,
                :cmd,
                :cwd,
                :stdout,
                :stderr,
                :retc,
                :start_time,
                :duration
            )
        '''
        output = output or ()
        rowid = self(q,
                     trun_id=trun_id,
                     cmd=json.dumps(cmd),
                     cwd=str(cwd),
                     stdout=b''.join(o.out for o in output if o.kind == 'output'),
                     stderr=b''.join(o.out for o in output if o.kind == 'error'),
                     retc=retc,
                     start_time=start_time.timestamp(),
                     duration=duration).lastrowid
        assert rowid is not None
        return ProcExecID(rowid)

    def iter_files(self, *, run_id: Optional[RunID] = None) -> Iterable[FileInfo]:
        """
        Iterate over every file in the database, optionally restricted to the
        given run.

        :param run_id: If not-``None``, only yields files attached to the given
            run.
        """
        if run_id is None:
            rows = self('SELECT file_id, run_id, path, size FROM dagon_sized_files')
        else:
            rows = self('SELECT file_id, run_id, path, size FROM dagon_sized_files WHERE run_id=:r', r=run_id)

        for fid, rid, path_, size in rows:
            yield FileInfo(run_id=rid, file_id=fid, path=PurePosixPath(path_), size=size)

    def iter_file_data(self, file: Union[FileInfo, FileID]) -> Iterable[bytes]:
        """
        Iterate over the chunks of data in the file stored in the database.

        :param file: The file to read from.
        """
        if isinstance(file, FileInfo):
            file = file.file_id
        for dat, in self('SELECT data FROM dagon_storage_file_data WHERE file_id=:fid ORDER BY nth', fid=file):
            yield dat

    def add_task_dep(self, *, run_id: RunID, dependent: str, depends_on: Optional[str]) -> None:
        """
        Store information about the dependencies between tasks.

        :param run_id: The run to attach this information to.
        :param dependent: The task that has the dependency.
        :param depends_on: The task which is being depended on.
        """
        self(
            '''
            INSERT INTO dagon_task_deps_rel (run_id, dependent, depends_on)
                 VALUES (:r, :d1, :d2)
            ''',
            r=run_id,
            d1=dependent,
            d2=depends_on,
        )

    def iter_task_deps(self, run_id: RunID) -> Iterable[tuple[str, str]]:
        """
        Iterate over the dependency information for tasks in a given run.
        """
        return cast(Iterable['tuple[str, str]'],
                    self('SELECT dependent, depends_on FROM dagon_task_deps_rel WHERE run_id=:r', r=run_id))


@fixup_dataclass_docs
@dataclasses.dataclass()
class _AppContext():
    db_path: Path | None = None


class GlobalContext(NamedTuple):
    database: Database
    run_id: RunID


class TaskContext(NamedTuple):
    task_id: TaskID
    task_run_id: TaskRunID
    start_time: datetime.datetime


class _TaskContextPriv(NamedTuple):
    pub: TaskContext
    iv_stack: list[IntervalID]


class _DatabaseExt(BaseExtension[_AppContext, GlobalContext, _TaskContextPriv]):
    dagon_ext_name = 'dagon.db'
    dagon_ext_requires = ['dagon.pools', 'dagon.events']

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        grp = arg_parser.add_argument_group('Task Database Options')
        grp.add_argument('--db-path',
                         metavar='<path>',
                         type=Path,
                         help='Path to the database file to use. Default is ".dagon.db"',
                         default=os.environ.get('DAGON_DATABASE_PATH', '.dagon.db'))

    def handle_options(self, opts: argparse.Namespace) -> None:
        self.app_data().db_path = opts.db_path

    def app_context(self) -> ContextManager[_AppContext]:
        return nullcontext(_AppContext())

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[GlobalContext]:
        db = Database.open(self.app_data().db_path or '.dagon.db')
        rid = db.new_run_id()
        # Yield now to let the graph run
        yield GlobalContext(db, rid)
        # Link up all the dependencies
        for task in graph.all_nodes:
            for dep in graph.dependencies_of(task):
                db.add_task_dep(run_id=rid, dependent=task.name, depends_on=dep.name)

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[_TaskContextPriv]:
        db = self.global_data()
        now = datetime.datetime.now()
        tid = db.database.get_task_rowid(task.name)
        trun_id = db.database.add_task_run(run=db.run_id, task=tid, state=TaskState.Running, start_time=now)
        with ExitStack() as st:
            st.enter_context(events['dagon.interval-start'].connect(self._iv_start))
            st.enter_context(events['dagon.interval-end'].connect(self._iv_end))
            st.enter_context(
                events.get_or_register('dagon.proc.done', Event[CompletedProcess]).connect(self._proc_done))
            yield _TaskContextPriv(TaskContext(tid, trun_id, now), [])

    def _iv_start(self, name: str) -> None:
        iv = self.global_data().database.new_interval(self.task_data().pub.task_run_id, name)
        self.task_data().iv_stack.append(iv)

    def _iv_end(self, _: None) -> None:
        iv = self.task_data().iv_stack.pop()
        self.global_data().database.set_interval_end(iv)

    def _proc_done(self, p: CompletedProcess) -> None:
        # Store this process execution in the database
        db = self.global_data().database
        trun_id = self.task_data().pub.task_run_id
        iv = db.new_interval(trun_id, f'Subprocess {p.command}', p.start_time)
        dur = p.end_time - p.start_time
        pid = db.store_proc_execution(
            trun_id,
            cmd=p.command,
            cwd=p.cwd,
            output=p.result.output,
            retc=p.result.retcode,
            start_time=p.start_time,
            duration=dur.total_seconds(),
        )
        meta = {'process_id': pid}
        db.set_interval_meta(iv, meta)
        db.set_interval_end(iv, p.end_time)

    async def notify_result(self, result: NodeResult[OpaqueTask]) -> None:
        db = self.global_data()
        tctx = self.task_data().pub
        st = TaskState.Pending
        if isinstance(result.result, Cancellation):
            st = TaskState.Cancelled
        elif isinstance(result.result, Failure):
            st = TaskState.Failed
        else:
            assert isinstance(result.result, Success)
            st = TaskState.Succeeded
        db.database.update_task_state(tctx.task_run_id, state=st, duration=datetime.datetime.now() - tctx.start_time)


def global_context_data() -> GlobalContext | None:
    try:
        return _DatabaseExt.global_data()
    except LookupError:
        return None


def task_context_data() -> TaskContext | None:
    try:
        return _DatabaseExt.task_data().pub
    except LookupError:
        return None



from __future__ import annotations

import random

import pytest

from .. import db as db_mod
from .. import persist as mod
from ..task.task_test import run_test_on_fun


@pytest.fixture
def db() -> db_mod.Database:
    return db_mod.Database.open(':memory:')


EGGS = mod.Key[int]('eggs')


@pytest.mark.asyncio
async def test_persist_global(db: db_mod.Database) -> None:
    with mod.set_global_persistor(mod.in_database(db)):
        assert not await mod.globl.has(EGGS)
        await mod.globl.set(EGGS, 22)
        assert await mod.globl.get(EGGS) == 22

    with mod.set_global_persistor(mod.in_database(db)):
        # Reopen the same persistor
        assert await mod.globl.has(EGGS)
        assert await mod.globl.get(EGGS) == 22

        # Modify it again
        await mod.globl.set(EGGS, 42)
        assert await mod.globl.get(EGGS) == 42

    with mod.set_global_persistor(mod.in_database(db)):
        assert await mod.globl.get(EGGS) == 42


@pytest.mark.asyncio
async def test_persist_local(db: db_mod.Database) -> None:
    tid1 = db.get_task_rowid('example1')
    tid2 = db.get_task_rowid('example2')

    with mod.set_task_local_persistor(mod.in_database(db, task=tid1)):
        assert not await mod.local.has(EGGS)
        await mod.local.set(EGGS, 12)
        assert await mod.local.get(EGGS) == 12

    with mod.set_task_local_persistor(mod.in_database(db, task=tid2)):
        assert not await mod.local.has(EGGS)
        await mod.local.set(EGGS, 72)
        assert await mod.local.get(EGGS) == 72

    with mod.set_task_local_persistor(mod.in_database(db, task=tid1)):
        # Restored from the prior task ID
        assert await mod.local.has(EGGS)
        assert await mod.local.get(EGGS) == 12


def test_task_persist_global() -> None:
    val = random.randint(0, 99999)

    tv = mod.Key[int]('test_value')

    async def run() -> None:
        await mod.globl.set(tv, val)

    async def run2() -> None:
        assert await mod.globl.get(tv) == val

    run_test_on_fun(run)
    run_test_on_fun(run2)
    run_test_on_fun(run)
    run_test_on_fun(run2)
    run_test_on_fun(run2)


def test_task_persist_local() -> None:
    val1 = random.randint(0, 5_000)
    val2 = random.randint(5_001, 10_000)

    did_run1 = False
    value = mod.Key[int]('value')

    async def run1() -> None:
        nonlocal did_run1
        if did_run1:
            assert await mod.local.get(value) == val1
        else:
            await mod.local.set(value, val1)
        did_run1 = True

    did_run2 = False

    async def run2() -> None:
        nonlocal did_run2
        if did_run2:
            assert await mod.local.get(value) == val2
            assert await mod.local.get(value) != val1
        else:
            await mod.local.set(value, val2)
        did_run2 = True

    run_test_on_fun(run1)
    assert did_run1
    run_test_on_fun(run1)

    run_test_on_fun(run2)
    assert did_run2
    run_test_on_fun(run2)



import pytest
from dagon import task

from tests.test_util import dag_test

from .. import cache
from ..cache import Key


@dag_test()
def test_simple_usage():
    foo_key = Key[int]('foo')

    @task.define()
    async def t():
        cache.task_run.set(Key[int]('foo'), 12)
        get = cache.task_run.get(Key[int]('foo'))
        assert get == 12

    @task.define()
    async def f1():
        with pytest.raises(KeyError):
            cache.task.get(foo_key)
        cache.task.set(foo_key, 12)
        with pytest.raises(KeyError):
            cache.graph.get(foo_key)
        cache.graph.set(foo_key, 54)

        with pytest.raises(LookupError):
            cache.graph.child('bar')
        child = cache.graph.child('bar', if_missing='create')
        with pytest.raises(KeyError):
            child.get(foo_key)
        child.set(foo_key, 42)

    @task.define(depends=[f1])
    async def f2():
        with pytest.raises(KeyError):
            cache.task.get(foo_key)
        assert cache.graph.get(foo_key) == 54
        assert cache.graph.child('bar').get(foo_key) == 42

    return t, f1, f2



"""
Module ``dagon.fs``
###################

Module for performing filesystem operations
"""

from __future__ import annotations

import asyncio
import errno
import itertools
import os
import shutil
import stat
from concurrent.futures import ThreadPoolExecutor
from io import BufferedIOBase
from pathlib import Path, PurePath
from typing import (AsyncContextManager, AsyncIterable, AsyncIterator, Awaitable, Callable, Iterable, Mapping, TypeVar,
                    Union, overload)

from typing_extensions import Literal

from .. import util
from ..event import CancellationToken, raise_if_cancelled
from ..util import T

Pathish = Union['os.PathLike[str]', str]
'A path-able argument. A string, subclass of `~pathlib.PurePath`, or anything with an ``__fspath__`` member'
NPaths = Union[Pathish, Iterable[Pathish]]
'Convenience type to represent either a single path or an iterable list of paths'
FilePredicate = Callable[[Path], bool]
'Type for predicate functions that accept a file path and return a `bool`'

_FS_OPS_POOL = ThreadPoolExecutor(4)

_PathT = TypeVar('_PathT', bound=Pathish)


@overload
def iter_pathish(p: NPaths) -> Iterable[Path]:
    ...


@overload
def iter_pathish(p: NPaths, *, fac: Callable[[Pathish], _PathT]) -> Iterable[_PathT]:
    ...


def iter_pathish(p: NPaths, *, fac: Callable[[Pathish], _PathT] = Path) -> Iterable[_PathT]:
    """
    Given an iterable of paths or a single path, yield either the single path
    or each of the paths given.
    """
    if isinstance(p, (str, PurePath)) or hasattr(p, '__fspath__'):
        yield fac(p)  # type: ignore
    else:
        yield from (fac(i) for i in p)  # type: ignore


IfExists = Literal['replace', 'fail', 'keep']
"""Policy for performing file operations where the destination already exists"""

IfDirectoryExists = Literal['replace', 'fail', 'keep', 'merge']
"""Policy for performing file operations where the destination directory already exists"""


def fs_thread_pool() -> ThreadPoolExecutor:
    """
    The thread pool that is used to execute filesystem operations
    """
    return _FS_OPS_POOL


def _run_fs_op(fn: Callable[[], T]) -> asyncio.Future[T]:
    loop = asyncio.get_event_loop()
    return loop.run_in_executor(fs_thread_pool(), fn)  # type: ignore


def _copy_file_no_overwrite(src: Path, dst: Path) -> None:
    with src.open('rb') as f:
        with dst.open('xb') as out:
            shutil.copyfileobj(f, out)


def _any_parent_is_file(p: Path) -> bool:
    "Determine whether 'p' (or any ancestor of 'p') names a file"
    if p.is_dir():
        return False
    if p.is_file():
        return True
    return _any_parent_is_file(p.parent)


async def copy_file(file: Pathish,
                    dest: Pathish,
                    *,
                    if_exists: IfExists = 'fail',
                    mkdirs: bool = True,
                    preserve_symlinks: bool = True,
                    preserve_stat: bool = False,
                    cancel: CancellationToken | None = None) -> Path:
    """
    Copy a file from one location to another.

    :param file: Path to a file to copy.
    :param dest: The full filepath for the file destination.
    :param on_exists: What to do if the destination already exists.
    :param mkdirs: If ``True``, make intermediate directories for the destination.
    :param preserve_symlinks: If ``True``, and ``file`` refers to a symlink, the
        symlink will be copied (rather than the target thereof)
    :param preserve_stat: If ``True``, attributes of the source file will be
        applied to the destination.
    :param cancel: A cancellation token to cancel the operation.

    .. note::
        ``dest`` is the full path to a file, not the path to the directory that
        should receive the file.
    """
    cancel = cancel or CancellationToken.get_context_local()
    raise_if_cancelled(cancel)
    file = Path(file)
    dest = Path(dest)
    pardir = dest.parent
    if not pardir.exists() and mkdirs:
        pardir.mkdir(exist_ok=True, parents=True)

    stat = file.stat()
    is_small = stat.st_size < 1024 * 64
    do_copy: Callable[[Path, Path], None]
    if preserve_symlinks and file.is_symlink():
        # Copy the link, not the target of the link
        do_copy = lambda s, d: os.symlink(os.readlink(s), d)
    else:
        # Copy file, but raise if we attempt to overwrite it
        do_copy = _copy_file_no_overwrite
    try:
        if is_small:
            do_copy(file, dest)
        else:
            await _run_fs_op(lambda: do_copy(file, dest))
    except FileNotFoundError as e:
        # Windows might not raise NotADirectoryError itself, so we'll check that
        # case ourselves
        if _any_parent_is_file(dest):
            raise NotADirectoryError from e
        raise
    except PermissionError as e:
        if file.is_dir():
            raise IsADirectoryError from e
        raise
    except FileExistsError:
        if if_exists == 'fail':
            raise
        if if_exists == 'keep':
            return dest
        assert if_exists == 'replace'
        # Delete the file, and do it again.
        dest.unlink()
        if is_small:
            do_copy(file, dest)
        else:
            await _run_fs_op(lambda: do_copy(file, dest))

    if preserve_stat:
        shutil.copystat(file, dest, follow_symlinks=False)
    else:
        shutil.copymode(file, dest, follow_symlinks=False)
    return dest


async def copy_tree(inpath: Pathish,
                    destpath: Pathish,
                    *,
                    mkdirs: bool = True,
                    if_exists: IfDirectoryExists = 'fail',
                    if_file_exists: IfExists = 'fail',
                    only_if: FilePredicate | None = None,
                    unless: FilePredicate | None = None,
                    preserve_symlinks: bool = True,
                    preserve_stat: bool = False,
                    cancel: CancellationToken | None = None) -> Path:
    """
    Copy a tree file files/directories.

    :param inpath: Path to a file/directory to copy.
    :param destpath: The destination path to copy to.
    :param mkdirs: If ``True``, create intermediate directories.
    :param on_exists: The action to take if the destination directory exists.
    :param on_file_exists: The action to take if any files within the destination
        directory already exists. (This is applicable to ``on_exists="merge"``)
    :param only_if: A predicate function that takes an absolute ``Path`` and
        returns a ``bool``. If provided, only files for which this predicate
        returns ``True`` will be copied into the destination.
    :param unless: A predictae function that takes an absolute ``Path`` and
        returns a ``bool``. If provided, only files for which this predicate
        returns ``False`` will be copied into the destination.
    :param preserve_symlinks: Whether to copy symlinks as symlinks, or follow
        them through to copy their targets.
    :param preserve_stat: If ``True``, attributes from the source will be
        applied to the destination.
    :param cancel: A cancellation token for the operation.

    .. note::
        If both ``only_if`` and ``unless`` are provided, ``only_if`` will be
        checked *first*, and then ``unless`` will be checked. Both predicates
        will be considered.
    """
    cancel = cancel or CancellationToken.get_context_local()
    raise_if_cancelled(cancel)
    indir = Path(inpath).absolute()
    destdir = Path(destpath)
    pardir = destdir.parent
    if not pardir.is_dir():
        if mkdirs:
            pardir.mkdir(exist_ok=True, parents=True)
        else:
            raise NotADirectoryError(f'Cannot copy directory {indir} to {destdir}: Path {pardir} '
                                     'is not an existing directory')
    if destdir.exists():
        if if_exists == 'replace':
            shutil.rmtree(destdir)
        if if_exists == 'fail':
            raise FileExistsError(f'Cannot copy {indir} to {destdir}: Destination already exists')
        if if_exists == 'keep':
            return destdir
        if if_exists == "merge":
            pass

    destdir.mkdir(exist_ok=True, parents=True)
    raise_if_cancelled(cancel)
    for in_abs in indir.iterdir():
        raise_if_cancelled(cancel)
        dest_abs = destdir / (in_abs.relative_to(indir))
        # Item is a symlink, and that symlink refers to a directory, and
        # we are not being asked to preserve symlinks. Resolve the link
        # and copy the tree from the directory:
        should_follow_dir_symlink = (in_abs.is_symlink() and not preserve_symlinks and in_abs.resolve().is_dir())
        # Whether we should copy the tree:
        should_copy_tree = in_abs.is_dir() or should_follow_dir_symlink
        if should_copy_tree:
            await copy_tree(in_abs,
                            dest_abs,
                            if_file_exists=if_file_exists,
                            only_if=only_if,
                            unless=unless,
                            preserve_stat=preserve_stat,
                            preserve_symlinks=preserve_symlinks,
                            cancel=cancel)
            # Process the next item
            continue

        # This is a file, not a directory
        if only_if is not None and not only_if(in_abs):
            # The file has been excluded by the `only_if` filter
            continue
        if unless is not None and unless(in_abs):
            # The file has been excluded by the `unless` filter
            continue
        await copy_file(in_abs,
                        dest_abs,
                        if_exists=if_file_exists,
                        mkdirs=False,
                        preserve_symlinks=preserve_symlinks,
                        preserve_stat=preserve_stat)
    if preserve_stat:
        shutil.copystat(indir, destdir, follow_symlinks=False)
    return destdir


TreeItem = Union[bytes, Mapping[str, 'TreeItem']]


async def create_tree(root: Pathish, items: Mapping[str, TreeItem], *, cancel: CancellationToken | None = None) -> None:
    cancel = cancel or CancellationToken.get_context_local()
    raise_if_cancelled(cancel)
    return await _run_fs_op(lambda: _create_fs_tree(root, items, cancel))


def _create_fs_tree(root: Pathish, items: Mapping[str, TreeItem], cancel: CancellationToken | None) -> None:
    root = Path(root)
    root.mkdir(exist_ok=True, parents=True)
    for filename, item in items.items():
        raise_if_cancelled(cancel)
        if isinstance(item, bytes):
            root.joinpath(filename).write_bytes(item)
        else:
            _create_fs_tree(root / filename, item, cancel)


_delete_increment = 0


def _remove1(f: Path, recurse: bool, absent_ok: bool, cancel: CancellationToken | None) -> Awaitable[None]:
    raise_if_cancelled(cancel)
    global _delete_increment
    _delete_increment += 1
    tmpname = f.parent / f'{f.name}.del-{_delete_increment}'
    try:
        f.rename(tmpname)
    except FileNotFoundError:
        if not absent_ok:
            raise
        # We're okay if this file is missing
        return util.ReadyAwaitable(None)
    if tmpname.is_dir():
        if not recurse:
            raise IsADirectoryError(f)
        return _run_fs_op(lambda: _remove_dir(tmpname, cancel))
    _remove_file_or_dir(tmpname)
    return util.ReadyAwaitable(None)


def _remove_file_or_dir(f: Path) -> None:
    try:
        f.unlink()
    except PermissionError:
        _make_writable_retry(os.remove, str(f))


def _remove_dir(dirpath: Path, cancel: CancellationToken | None) -> None:
    raise_if_cancelled(cancel)
    shutil.rmtree(dirpath, onerror=_make_writable_retry)


def _make_writable_retry(
    fn: Callable[[str], None],
    path: str,
    *_ignore: None,
) -> None:
    if fn is os.chmod:
        raise
    os.chmod(path, stat.S_IWRITE)
    fn(path)


def remove(files: NPaths,
           *,
           recurse: bool = False,
           absent_ok: bool = False,
           cancel: CancellationToken | None = None) -> Awaitable[None]:
    """
    Remove one or more files or directories.

    :param files: Path(s) to files and/or directories to remove.
    :param recurse: If ``True`` and any of ``files`` names a directory,
        recursively remove the directory and all of its contents.
    :param absent_ok: If ``True`` and any named file does not exist, no error
        will be raised. This can create simple idempotent file/directory
        removal.
    :param cancel: A cancellation token for the operation.

    .. hint::

        This "remove" function is optimized for getting files "out of the way"
        as fast as possible. As soon as this function returns (even before the
        return value is ``await``\\ ed), all named files/directories will be
        gone from their original location.

        Before being deleted from disk, the named files/directories are moved to
        a new temporary name. This move operation is extremely fast, whereas a
        recursive directory deletion might be extremely slow. For this reason,
        you can call :func:`remove` without awaiting, then act "as if" the files
        are already deleted (they are moved to a temporary location and being
        deleted in the background). Await the result of :func:`remove` to block
        until the deletion operations are actually completed.
    """
    cancel = cancel or CancellationToken.get_context_local()
    if isinstance(files, (str, PurePath)):
        return _remove1(Path(files), recurse, absent_ok, cancel)
    multi_files = iter_pathish(files)
    multi_rm = map(lambda f: _remove1(f, recurse, absent_ok, cancel), multi_files)
    futs = set(map(asyncio.ensure_future, multi_rm))
    return _remove_gather(futs)


async def _remove_gather(futs: set[asyncio.Task[None]]) -> None:
    done, pending = await asyncio.wait(futs, return_when=asyncio.FIRST_EXCEPTION)
    for t in pending:
        # 'pending' is non-empty iff an exception occurred. Cancel all other
        # running operations.
        t.cancel()
    for f in done:
        await f
    assert len(done) == len(futs)


def clear_directory(dirs: NPaths, *, cancel: CancellationToken | None = None) -> Awaitable[None]:
    """
    Remove the contents of one or more directories, but keep the directories.

    :param dirs: Path(s) to an directories to clear.
    :param cancel: A cancellation token for the operation.
    """
    each_dir = iter_pathish(dirs)
    children_per_dir = (d.iterdir() for d in each_dir)
    all_children = itertools.chain.from_iterable(children_per_dir)
    return remove(all_children, recurse=True, absent_ok=True, cancel=cancel)


def _read_some(fd: BufferedIOBase, buf: bytearray) -> int:
    return fd.readinto(buf)


async def read_blocks_from(io: BufferedIOBase,
                           *,
                           block_size: int = 1024 * 10,
                           cancel: CancellationToken | None = None) -> AsyncIterator[memoryview]:
    buf = bytearray(block_size)
    cancel = CancellationToken.get_context_local()
    while True:
        raise_if_cancelled(cancel)
        nread = await _run_fs_op(lambda: _read_some(io, buf))
        if nread == 0:
            break
        yield memoryview(buf)[:nread]


async def read_blocks(filepath: Pathish,
                      *,
                      block_size: int = 1024 * 10,
                      cancel: CancellationToken | None = None) -> AsyncIterable[memoryview]:
    """
    Asynchronously lazily ready blocks from the given file.
    """
    raise_if_cancelled(cancel)
    filepath = Path(filepath)
    with filepath.open('rb') as fd:
        async for b in read_blocks_from(fd, block_size=block_size, cancel=cancel):
            yield b


def safe_move_file(source: Pathish,
                   dest: Pathish,
                   *,
                   mkdirs: bool = True,
                   if_exists: IfExists = 'fail',
                   cancel: CancellationToken | None = None) -> Awaitable[Path]:
    """
    Move the given file from one location to another. If the system cannot
    simply move the file, it will be copied and the source will be removed.

    :param source: The source file which will be moved.
    :param dest: The destination of the move operation.
    :param mkdirs: If ``True``, the directory containing ``dest`` will be created if absent.
    :param on_exists: What to do if the destination already exists.
    :param cancel: A cancellation token for the operation.
    """
    source = Path(source)
    dest = Path(dest)
    cancel = cancel or CancellationToken.get_context_local()
    raise_if_cancelled(cancel)
    if mkdirs:
        dest.parent.mkdir(parents=True, exist_ok=True)
    try:
        source.rename(dest)
        return util.ReadyAwaitable(dest)
    except OSError as e:
        # Check if the destination file already exists
        if e.errno == errno.EEXIST:
            if if_exists == 'fail':
                # User wants to receive this exception
                raise
            if if_exists == 'keep':
                # User wants to keep the destination file
                return util.ReadyAwaitable(dest)
            # The alternative is to replace
            c: Literal['replace'] = if_exists
            util.unused(c)
            # User wants us to replace the file in the destination
            _remove_file_or_dir(dest)
            return safe_move_file(source, dest, mkdirs=mkdirs, if_exists=if_exists, cancel=cancel)
        if e.errno == errno.EXDEV:
            # Cross-device linking: We're trying to relink a file across a filesystem
            # boundary. This is where the "safe" comes in.
            return _teleport_file(source, dest, cancel)
        raise


async def _teleport_file(source: Path, dest: Path, cancel: CancellationToken | None) -> Path:
    await copy_file(file=source, dest=dest, cancel=cancel, preserve_stat=True, preserve_symlinks=True)
    _remove_file_or_dir(source)
    return dest


def open_file_writer(fpath: Pathish, if_exists: IfExists = 'fail'):
    """
    Opens a native file as an `~IFileWriter`.
    """
    from ..storage import NativeFileStorage
    return NativeFileStorage('/').open_file_writer(fpath, if_exists=if_exists)


def read_file(fpath: Pathish) -> AsyncContextManager[AsyncIterable[bytes]]:
    """
    Opens a native file as an asyncronous byte iterable.
    """
    from ..storage import NativeFileStorage
    return NativeFileStorage('/').read_file(fpath)



from datetime import datetime
from pathlib import Path

import pytest

from .. import db as mod


def get_mem_db() -> mod.Database:
    return mod.Database.open(':memory:')


# Mark tests as asyncio to ensure there is a thread loop for the database
def test_create() -> None:
    mod.open_sqlite_db(':memory:')


def test_new_run_id() -> None:
    db = get_mem_db()
    run_id = db.new_run_id()
    assert run_id == 1
    run_id_2 = db.new_run_id()
    assert run_id_2 == 2


def test_store_proc_exec() -> None:
    db = get_mem_db()
    run = db.new_run_id()
    t = db.get_task_rowid('test')
    trun_id = db.add_task_run(run=run, task=t)
    db.store_proc_execution(
        trun_id,
        cmd=['foo', 'bar'],
        cwd=Path('/nowhere'),
        output=(),
        retc=12,
        start_time=datetime.now(),
        duration=12,
    )



from datetime import datetime, timedelta
from pathlib import Path
import pytest

from ..util.testing import async_test
from . import _cache as mod


def test_parse_cc():
    cc = mod.CacheControl.parse('no-store, max-age=124')
    assert cc.no_store
    assert not cc.must_revalidate
    assert cc.max_age == 124


def test_mangle_path():
    p = mod.cachepath_for_url('http://google.com', 31)
    assert p == 'http/google.com-nil/c7b9/_-31'

    p = mod.cachepath_for_url('http://google.com/', 8)
    assert p == 'http/google.com-nil/6ab0/_-8'

    p = mod.cachepath_for_url('http://google.com/foo/bar', 0)
    assert p == 'http/google.com-nil/af3e/foo/bar-0'

    p = mod.cachepath_for_url('http://google.com/foo/bar/', 91)
    assert p == 'http/google.com-nil/6f68/foo/bar-91'


@async_test
async def test_save(tmp_path: Path):
    c = mod.open_cache_in(tmp_path)
    f = tmp_path.joinpath('eggs.txt')
    f.write_text('hello, cache!')
    dltime = datetime.now()
    await c.save(url='https://google.com',
                 last_modified=None,
                 etag='meow',
                 cache_control='max-age=30',
                 download_time=dltime,
                 age=12,
                 take_file=f)

    avail = list(c.for_url('https://google.com'))
    assert avail == [
        mod.CacheEntry(
            1,
            'https://google.com',
            'meow',
            mod.NIL_CACHE_CONTROL._replace(max_age=30),
            None,
            12,
            dltime,
        )
    ]
    assert c.path_to(avail[0]).read_text() == 'hello, cache!'
    first = avail[0]
    assert first.age_offset == 12
    assert first.etag == 'meow'
    assert first.expires_at == dltime + timedelta(seconds=30 - 12)
    assert not first.is_stale



"""
Module ``dagon.option``
#######################

Create run-time execution options to customize task behaviors.

Task can define customization options using this API to tweak task behaviors.
The value assigned to any option is fixed at task graph invocation and is
uniform across a task graph execution.

Using `.option.add` to create new options while declaring tasks, and use
`.option.Option.get` or `.option.value_of` to obtain the value provided to an
option at runtime.

.. data:: OptionType
    :canonical: dagon.option.option.OptionType
    :type: str | bool | pathlib.Path | int | float | enum.Enum

    A type union of any type that can be used with `.Option`.

    .. _option-types:

    Dagon supports several types by default:

    - `str` for strings.
    - `bool` for true/false values.
    - :class:`pathlib.Path` to represent filesystem paths.
    - `int` for integral values.
    - `float` for floating point numbers.
    - Any subclass of :class:`enum.Enum`. It is recommended to use `str` keys
      with enums when using such types.

    In addition, Dagon supports any class that presents a `__dagon_parse_opt__`
    static method, which should receive a string and return an instance of the
    class.

.. data:: OptionT
    :canonical: dagon.option.option.OptionT

    A type variable bound to `.OptionType`.
"""

from __future__ import annotations

import argparse
import enum
import re
import sys
import textwrap
from contextlib import nullcontext
from dataclasses import dataclass, field
from inspect import isclass
from pathlib import Path
from typing import (TYPE_CHECKING, Any, Callable, ContextManager, Generic, Iterable, Mapping, Sequence, Type, TypeVar,
                    Union, cast, overload)
from typing_extensions import Protocol

import dagon.tool.args

from ..ext.base import BaseExtension
from ..util import DefaultSentinelType, T, U, typecheck

__all__ = ['Option', 'add', 'NoOptionValueError']


class _CustomDagonOpt(Protocol):
    @classmethod
    def __dagon_parse_opt__(cls, val: str) -> _CustomDagonOpt:
        ...


SimpleOptionType = Union[str, bool, Path, int, float, enum.Enum, _CustomDagonOpt]

SimpleT = TypeVar('SimpleT', bound=SimpleOptionType)

SimpleOptionT = TypeVar('SimpleOptionT', bound=SimpleOptionType)


def convert_bool(s: str) -> bool:
    """
    Convert the given string into a bool with some common bool-ish values.
    """
    b = {
        '0': False,
        'no': False,
        'n': False,
        'false': False,
        '1': True,
        'yes': True,
        'y': True,
        'true': True,
    }.get(s.lower())
    if b is None:
        raise ValueError(f'Invalid boolean option string "{s}"')
    return b


_CONVERSION_MAP: Mapping[Type[Any], Callable[[str], Any]] = {
    str: str,
    bool: convert_bool,
    Path: Path,
    int: int,
    float: float,
}

_PARSE_OPT_METHOD_KEY = '__dagon_parse_opt__'


class NoSuchOptionError(NameError):
    pass


def is_valid_option_type(typ: Type[Any]) -> bool:
    """
    Return `True` if the given type is valid for use as a type with :class:`Option`.
    """
    return hasattr(typ, _PARSE_OPT_METHOD_KEY) or issubclass(typ, enum.Enum) or (typ in _CONVERSION_MAP)


def parse_value_str(typ: Type[SimpleOptionT], val_str: str) -> SimpleOptionT:
    """
    Given a type `typ` and a value string `val_str`, attempt to convert
    that string to the given `typ`. Raises `ValueError` in cases of
    failure.
    """
    try:
        conv = getattr(type, _PARSE_OPT_METHOD_KEY, None)
        if conv is not None:
            assert callable(conv)
            return cast(SimpleOptionT, conv(val_str))  # pylint: disable=not-callable
        if issubclass(typ, enum.Enum):
            try:
                return cast(SimpleOptionT, typ(val_str))
            except ValueError as e:
                try:
                    i = int(val_str)
                except ValueError:
                    raise e from e
                else:
                    return cast(SimpleOptionT, typ(i))
        return cast(SimpleOptionT, _CONVERSION_MAP[typ](val_str))
    except Exception as e:
        raise ValueError(f'Failed to parse "{val_str}" as type `{typ.__name__}`') from e


def get_type_parser(typ: Type[SimpleOptionT]) -> Callable[[str], SimpleOptionT]:
    return lambda s: parse_value_str(typ, s)


_DEFAULT_SENTINEL: Any = DefaultSentinelType()


class Option(Generic[T], Protocol):
    """
    Representation of a run-time execution option that can be provided by a
    user.
    """
    @property
    def name(self) -> str:
        ...

    @property
    def doc(self) -> str | None:
        ...

    @property
    def type(self) -> Type[T]:
        ...

    @overload
    def get(self) -> T:
        ...

    @overload
    def get(self, *, default: U) -> T | U:
        ...

    @overload
    def get(self, *, default_factory: Callable[[], U]) -> T | U:
        ...


class _Option(Generic[T]):
    """
    Representation of a run-time execution option that can be provided by a
    user.

    :param name: The name of the option. Should be unique in a given Dag.
    :param typ: The type of the option. Must be a subtype of `.OptionType`.
    :param calc_default: The default value, or a factory function that generates
        the default value. If omitted and no value was provided by a user,
        requesting the option's value will raise an exception.
    :param doc: A documentation string to display to the user.
    :param validate: A secondary validation function that checks that a given
        value is valid for this option. Should return `None` when the value
        is valid, or an error string explaining why it is an invalid value.
    """
    def __init__(self,
                 name: str,
                 *,
                 type: Type[T],
                 parse: Callable[[str], T],
                 calc_default: Callable[[], T] | None = None,
                 doc: str | None = None,
                 validate: Callable[[T], str | None] | None = None) -> None:
        self._name = name
        self._parser = parse
        self._type = type
        self._calc_default = calc_default
        self._validate = validate
        self._doc = doc

    @property
    def name(self) -> str:
        """The option's name"""
        return self._name

    @property
    def doc(self) -> str | None:
        """Documentation for this option"""
        return self._doc

    def get(self, *, default: U = _DEFAULT_SENTINEL, default_factory: Callable[[], U] = _DEFAULT_SENTINEL) -> T | U:
        """
        Obtain the value that was given to this option.

        .. note::
            This method may only be called from within a task-executing context!
        """
        return _ctx_fulfilled_options().get(self, default=default, default_factory=default_factory)

    @property
    def default_factory(self) -> None | Callable[[], T]:
        """Obtain the (optional) default-value factory for this option"""
        return self._calc_default

    def get_default_or_raise(self, default_value: T, default_factory: Callable[[], U]) -> T | U:
        if self._calc_default is not None:
            return self._calc_default()
        if default_value is not _DEFAULT_SENTINEL:
            return default_value
        if default_factory is not _DEFAULT_SENTINEL:
            return default_factory()
        raise NoOptionValueError(f'No value was provided for option "{self.name}", and no default value was specified')

    @property
    def type(self) -> Type[T]:
        return self._type

    def validate(self, value: T) -> str | None:
        """
        Check if the given value is valid for this option.

        Returns a string representing the error, or `None` if there was
        no error.
        """
        if self._validate is not None:
            return self._validate(value)
        return None

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: Any) -> bool:
        return self is other

    def __repr__(self) -> str:
        return f'<dagon.option.Option "{self.name}">'

    def parse_str(self, val_str: str) -> T:
        """
        Parse the given string into a value for this option. Does not perform
        additional validation.
        """
        return self._parser(val_str)


if TYPE_CHECKING:
    typecheck(Option[Any])(_Option[Any])


@overload
def add(name: str,
        type: Type[SimpleT],
        *,
        doc: str = '',
        validate: Callable[[SimpleT], str | None] = ...) -> Option[SimpleT]:
    ...


@overload
def add(name: str,
        type: Type[SimpleT],
        *,
        doc: str = '',
        validate: Callable[[SimpleT], str | None] = ...,
        default: U) -> Option[SimpleT | U]:
    ...


@overload
def add(name: str,
        type: Type[SimpleT],
        *,
        doc: str = '',
        validate: Callable[[SimpleT], str | None] = ...,
        default_factory: Callable[[], U]) -> Option[SimpleT | U]:
    ...


@overload
def add(name: str, type: Type[T] = ..., *, doc: str = '', parse: Callable[[str], T]) -> Option[T]:
    ...


@overload
def add(name: str, type: Type[T] = ..., *, doc: str = '', parse: Callable[[str], T], default: U) -> Option[T | U]:
    ...


@overload
def add(name: str,
        type: Type[T] = ...,
        *,
        doc: str = '',
        parse: Callable[[str], T],
        default_factory: Callable[[], U]) -> Option[T | U]:
    ...


def add(name: str,
        type: Type[T] | None = None,
        *,
        parse: Callable[[str], T] | None = None,
        doc: str = '',
        validate: Callable[[Any], str | None] | None = None,
        default: Any = _DEFAULT_SENTINEL,
        default_factory: Any = _DEFAULT_SENTINEL) -> Option[Any]:
    """
    Add an option to the current task graph with the given name and type.
    """
    oset = _ctx_option_set()
    if type is not None:
        parse = cast(Callable[[str], T], get_type_parser(cast(Type[SimpleOptionType], type)))
    assert parse, 'option.add requires a "parse" function or a supported "type"'
    if default_factory is not _DEFAULT_SENTINEL:
        # The default is a callable
        return oset.add(_Option[Any](name,
                                     type=type,
                                     parse=parse,
                                     doc=doc,
                                     validate=validate,
                                     calc_default=default_factory))
    elif default is not _DEFAULT_SENTINEL:
        return oset.add(_Option[Any](name,
                                     type=type,
                                     parse=parse,
                                     doc=doc,
                                     validate=validate,
                                     calc_default=lambda: default))
    else:
        # No default value provided
        return oset.add(_Option[Any](name, type=type, parse=parse, doc=doc, validate=validate))


class NoOptionValueError(ValueError):
    pass


class _FulfilledOptions:
    """
    Type that holds the result of options that have been fulfilled.

    :param values: Iterable of pairs that define the values for each option.
        The value should match the type of the option, or be ``Undefined``, in
        which case asking for the option value will return ``None``
    """
    def __init__(self, values: Iterable[tuple[_Option[Any], Any]]):
        self._values = dict(values)

    def get(self, option: _Option[T], default: U, default_factory: Callable[[], U]) -> T | U:
        """
        Get the value of the given option.

        If the option exists but no value was provided, returns ``None``. If the
        option does not exist, raises ``KeyError``.
        """
        if not option in self._values:
            raise KeyError(f'No known option "{option.name}"')
        mine = self._values[option]
        if mine is not _DEFAULT_SENTINEL:
            return mine

        fac = option.default_factory
        if fac is not None:
            return fac()
        if default is not _DEFAULT_SENTINEL:
            return default
        if default_factory is not _DEFAULT_SENTINEL:
            return default_factory()
        raise NoOptionValueError(
            f'No value was provided for option "{option.name}", and no default value was specified')


class _OptionSet:
    """
    A collection of options.

    :param options: Initial options to populate with
    """
    def __init__(self, options: Iterable[_Option[Any]] = ()):
        self._options: dict[str, _Option[Any]] = {}
        for o in options:
            self.add(o)

    def add(self, opt: _Option[T]) -> _Option[T]:
        """
        Register the given option with this set. Returns the given option.

        Raises :class:`RuntimeError` if the given option name is already taken.
        """
        if opt.name in self._options:
            raise RuntimeError(f'Duplicate option "{opt.name}"')
        self._options[opt.name] = opt
        return opt

    def get(self, key: str) -> _Option[Any] | None:
        """Get an option by name, or ``None`` if it does not exist in this set."""
        return self._options.get(key)

    @property
    def options(self) -> Iterable[_Option[Any]]:
        """The options in this set"""
        return self._options.values()

    def _iter_fulfill(self, kvs: Iterable[str]) -> Iterable[tuple[_Option[Any], Any]]:
        opt_re = re.compile(r'(.+?)=(.*)')
        all_keys = set(self._options.keys())
        for spec in kvs:
            mat = opt_re.match(spec)
            if not mat:
                raise ValueError(f'Invalid option specifier "{spec}" (should bey <key>=<value>)')
            key, val_str = mat.groups()
            opt = self.get(key)
            if opt is None:
                raise NoSuchOptionError(key)
            try:
                value = opt.parse_str(val_str)
            except Exception as e:
                raise ValueError(f'Failed to parse given string "{val_str}" as a value for option "{opt.name}"') from e
            err = opt.validate(value)
            if err is not None:
                raise RuntimeError(f'Invalid value for `{key}`: {err}')
            yield opt, value
            all_keys.remove(opt.name)

        for key in all_keys:
            opt = self.get(key)
            assert opt
            yield opt, _DEFAULT_SENTINEL

    def fulfill(self, kvs: Iterable[str]) -> _FulfilledOptions:
        """
        Fulfill the options in this set using the list of strings, which should
        be of the format ``<key>=<value>``.

        If a given ``key`` does not correpsond to any option, raises
        ``NameError``. If an option was not given a value in ``kvs``, that
        option will return ``None`` in the returned :class:`FulfilledOptions`.
        """
        return _FulfilledOptions(self._iter_fulfill(kvs))


class _OptionsArgs(dagon.tool.args.ParsedArgs, Protocol):
    opts: Sequence[str] | None
    list_options: bool


@dataclass()
class _OptionsContext:
    options: _OptionSet = field(default_factory=_OptionSet)
    fulfilled: None | _FulfilledOptions = None


def _ctx_option_set() -> _OptionSet:
    return _OptionsExt.app_data().options


def _ctx_fulfilled_options() -> _FulfilledOptions:
    c = _OptionsExt.app_data()
    if c.fulfilled is None:
        raise RuntimeError('Options have not yet been fulfilled')
    return c.fulfilled


def _format_docstr(s: str) -> Iterable[str]:
    paragraphs = textwrap.dedent(s).split('\n\n')
    for par in paragraphs:
        for line in textwrap.wrap(par):
            yield line.strip()
        yield ''  # Empty line after paragraph


def _option_help_annot(t: Type[Any]) -> str:
    assert isclass(t), t
    if issubclass(t, enum.Enum):
        return f'{t.__name__} {{' + ', '.join(str(f.value) for f in t) + '}'
    return t.__name__


def _list_options(opts: _OptionSet, with_docs: bool = True) -> int:
    for opt in opts.options:
        print(f'{opt.name}: {_option_help_annot(opt.type)}')
        if with_docs and opt.doc:
            for line in _format_docstr(opt.doc):
                print(f'    {line}')
    return 0


class _OptionsExt(BaseExtension[_OptionsContext, None, None]):
    dagon_ext_name = 'dagon.options'

    def __init__(self) -> None:
        self._set = _OptionSet()

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        grp = arg_parser.add_argument_group('Task Options')
        grp.add_argument('-o',
                         '--opt',
                         dest='opts',
                         metavar='<key>=<value>',
                         action='append',
                         help='Specify the value for a task option (can be repeated)')
        grp.add_argument('--list-options', '-lo', action='store_true', help='List available task options and exit')

    def handle_options(self, opts: argparse.Namespace) -> None:
        args = cast(_OptionsArgs, opts)
        if args.list_options:
            _list_options(_ctx_option_set(), with_docs=not args.no_doc)
            sys.exit(0)
        self.app_data().fulfilled = _ctx_option_set().fulfill(args.opts or ())

    def app_context(self) -> ContextManager[_OptionsContext]:
        return nullcontext(_OptionsContext())



from pathlib import Path
import os
import platform

_SYSTEM = platform.system()


def user_home() -> Path:
    return Path('~').expanduser()


def user_config_path() -> Path:
    e = os.environ.get('XDG_CONFIG_HOME')
    if e:
        return Path(e)
    if _SYSTEM == 'Windows':
        e = os.environ.get('AppData')
        if e:
            return Path(e)
    if _SYSTEM == 'Darwin':
        return user_home() / 'Library/Preferences'
    return user_home() / '.config'


def user_cache_path() -> Path:
    e = os.environ.get('XDG_CACHE_HOME')
    if e:
        return Path(e)
    if _SYSTEM == 'Windows':
        e = os.environ.get('LocalAppData')
        if e:
            return Path(e)
    if _SYSTEM == 'Darwin':
        return user_home() / 'Library/Caches'
    return user_home() / '.cache'



"""
Module ``dagon.tool.main``
##########################
"""

from __future__ import annotations

import argparse
import asyncio
import importlib
import importlib.util
import logging
import os
import shlex
import subprocess
import sys
import textwrap
import traceback
from typing_extensions import Protocol
import warnings
from pathlib import Path
from typing import Iterable, NoReturn, Optional, Sequence, cast

import pkg_resources

from ..core.result import Cancellation, Failure
from ..ext.exec import ExtAwareExecutor
from ..ext.loader import ExtLoader
from ..task.dag import TaskDAG, populate_dag_context
from ..task.task import DisabledTaskError, InvalidTask
from ..tool.args import ParsedArgs, get_argparser

_g_loading_dag = False


def _format_docstr(s: str) -> Iterable[str]:
    paragraphs = textwrap.dedent(s).split('\n\n')
    for par in paragraphs:
        for line in textwrap.wrap(par):
            yield line.strip()
        yield ''  # Empty line after paragraph


def _list_tasks(dag: TaskDAG, with_docs: bool = True) -> int:
    for t in dag.tasks:
        name_line = t.name
        if t.depends:
            name_line += ' : ' + ' '.join(shlex.quote(dep.dep_name) for dep in t.depends)
        print(name_line)
        if with_docs and t.doc:
            for line in _format_docstr(t.doc):
                print(f'    {line}')
        if t.is_disabled:
            print(f'    [DISABLED: {t.disabled_reason}]')
    return 0


def _configure_logging(level: int) -> None:
    logging.basicConfig(level=level, format='[dagon] %(message)s')


def _print_failure(task_name: str, fail: Failure) -> None:
    if isinstance(fail.exception, asyncio.CancelledError):
        # Don't print exceptions which are just task cancellations
        return

    head = f'Task "{task_name}" FAILED'

    term_width = int(os.environ.get('COLUMNS', '80'))

    if isinstance(fail.exception, subprocess.CalledProcessError):
        print(f'{head}: Subprocess execution returned non-zero', file=sys.stderr)
        traceback.print_exception(fail.exception_type, fail.exception, fail.traceback)
        print(f'{" Process output ":=^{term_width}}', file=sys.stderr)
        print((fail.exception.stdout or b'').decode(), file=sys.stderr, end='')
        print((fail.exception.stderr or b'').decode(), file=sys.stderr, end='')
        print(f'{" End Output ":^^{term_width}}', file=sys.stderr)
    else:
        print(f'{head}: Encountered an exception during execution:', file=sys.stderr)
        traceback.print_exception(fail.exception_type, fail.exception, fail.traceback)


def run_with_args(dag: TaskDAG, exts: ExtLoader, args: ParsedArgs, default_tasks: Optional[Sequence[str]]) -> int:
    if args.list_tasks:
        return _list_tasks(dag, with_docs=not args.no_doc)

    _configure_logging(logging.DEBUG if args.debug else logging.INFO)

    tasks = list(args.tasks or default_tasks or (t.name for t in dag.tasks if t.is_default))
    if not tasks:
        warnings.warn('No tasks are set to be executed')

    try:
        ll_dag = dag.low_level_graph(tasks)
    except DisabledTaskError as e:
        print(f'Unable to create the execution plan: {e}', file=sys.stderr)
        return 1
    except InvalidTask as e:
        if e.candidate:
            print(f'No such task "{e.key}". Did you mean "{e.candidate.name}" ?', file=sys.stderr)
        else:
            print(f'No such task "{e.key}".', file=sys.stderr)
        return 1

    exe = ExtAwareExecutor(exts, ll_dag, catch_signals=True, fail_cancels=args.fail_cancels)

    results = exe.run_all_until_complete()
    failed = False
    for item in results.values():
        if isinstance(item.result, Failure):
            _print_failure(item.task.name, item.result)
            failed = True
        elif isinstance(item.result, Cancellation):
            print(f'Task "{item.task.name}" was cancelled')
            failed = True
        else:
            # Nothing to print for successes
            pass

    return 1 if failed else 0


def run_for_dag(dag: TaskDAG,
                exts: ExtLoader,
                *,
                argv: Optional[Sequence[str]] = None,
                default_tasks: Optional[Sequence[str]] = None) -> int:
    argv = argv if argv is not None else sys.argv[1:]
    parser = get_argparser(exts)
    args = cast(ParsedArgs, parser.parse_args(argv))
    exts.handle_options(cast(argparse.Namespace, args))
    return run_with_args(dag, exts, args, default_tasks)


def get_extensions() -> ExtLoader:
    return ExtLoader.default()


class MainArgParseResults(ParsedArgs, Protocol):
    file: Optional[str]
    module: Optional[str]
    dir: Optional[Path]
    version: bool


def main(argv: Sequence[str]) -> int:
    exts = get_extensions()
    with exts.app_context():
        return _main(argv, exts)


def _main(argv: Sequence[str], exts: ExtLoader) -> int:
    parser = get_argparser(exts)
    parser.add_argument('-f', '--file', help='The file that defines the tasks')
    parser.add_argument('-m', '--module', help='The module to import to define the tasks')
    parser.add_argument('--dir',
                        help='Add the given directory to the import/search path for finding tasks',
                        default=Path.cwd(),
                        type=Path)
    parser.add_argument(
        '--version',
        help='Display the version and exit',
        action='store_true',
    )
    args = cast(MainArgParseResults, parser.parse_args(argv))

    if args.version:
        version = pkg_resources.get_distribution('dagon').version
        print(version)
        return 0

    def_file = Path(args.file).absolute() if args.file else None
    def_mod: Optional[str] = args.module

    if def_mod and def_file:
        print('`--module` and `--file` cannot be used together', file=sys.stderr)
        return 2

    if not def_mod and not def_file:
        def_mod = 'dag'

    global _g_loading_dag  # pylint: disable=global-statement
    _g_loading_dag = True
    prev_path = list(sys.path)
    try:
        sys.path.insert(0, str(args.dir))
        dag = TaskDAG('<main>')
        if def_mod:
            with populate_dag_context(dag):
                try:
                    if def_mod in sys.modules:
                        importlib.reload(sys.modules[def_mod])
                    else:
                        importlib.import_module(def_mod)
                except ModuleNotFoundError:
                    traceback.print_exc(file=sys.stderr)
                    print(f'Failed to import the "{def_mod}" module for task definitions', file=sys.stderr)
                    return 1
        else:
            spec = importlib.util.spec_from_file_location('<dag-definition>', str(def_file))
            if not spec:
                print('Unable to find task definitions', file=sys.stderr)
                return 1
            mod = importlib.util.module_from_spec(spec)
            with populate_dag_context(dag):
                spec.loader.exec_module(mod)  # type: ignore
    finally:
        _g_loading_dag = False
        sys.path = prev_path

    exts.handle_options(cast(argparse.Namespace, args))
    return run_with_args(dag, exts, args, default_tasks=[])


# def execute_default(*, dag: TaskDAG | None = None, default: Sequence[str] | None = None) -> None:
#     if _g_loading_dag:
#         warnings.warn(
#             'Call to execute_default() while running main Dagon executable. execute_default() will do nothing.')
#         return
#     dag = dag or current_dag()
#     rc = run_for_dag(dag, argv=sys.argv[1:], default_tasks=default or [])
#     sys.exit(rc)


def start() -> NoReturn:
    sys.exit(main(sys.argv[1:]))



from __future__ import annotations

import asyncio
import contextvars
import inspect
import itertools
import os
import signal
import traceback
import warnings
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from types import FrameType
from typing import (TYPE_CHECKING, Any, AsyncIterator, Awaitable, Callable, Iterable, Iterator, Mapping, NamedTuple,
                    Sequence, TypeVar, cast, overload)

from typing_extensions import Literal, TypeAlias

from dagon.core.result import Failure

from ..core import LowLevelDAG, NodeResult, Success, exec
from ..util import (NoneSuch, Opaque, first, nearest_matching, scope_set_contextvar)
from .task import DisabledTaskError, InvalidTask, Task

if TYPE_CHECKING:
    from dagon.event.cancel import CancellationToken

T = TypeVar('T')

OpaqueTask = Task[Opaque]


class NoDependencyError(RuntimeError):
    def __init__(self, from_task: Task[Any], to_task: Task[Any]):
        super().__init__(f'Task "{from_task.name}" does not have a direct dependency on "{to_task.name}"')
        self.from_task = from_task
        self.to_task = to_task


class IsOrderOnlyDependencyError(RuntimeError):
    def __init__(self, from_task: Task[Any], to_task: Task[Any]) -> None:
        super().__init__(f'The dependcy of "{from_task.name}" on "{to_task.name}" is an order-only dependency')
        self.from_task = from_task
        self.to_task = to_task


async def default_execute(task: Task[T]) -> T:
    return await task.function()


class TaskDAG:
    def __init__(self, name: str) -> None:
        self._name = name
        self._tasks: dict[str, OpaqueTask] = {}

    @property
    def tasks(self) -> Iterable[Task[Any]]:
        """Tasks associated with this DAG"""
        return self._tasks.values()

    def add_task(self, task: Task[T]) -> Task[T]:
        if task.name in self._tasks:
            raise RuntimeError(f'More than one task with name "{task.name}"')
        for dep in task.depends:
            if dep.dep_name not in self._tasks.keys():
                # pylint: disable=cell-var-from-loop
                cand = nearest_matching(dep.dep_name, self.tasks, key=lambda t: t.name)
                raise InvalidTask(NoneSuch[OpaqueTask](dep.dep_name, cand))
        self._tasks[task.name] = cast(OpaqueTask, task)
        return task

    def _collect(self, mark: str, stack: list[str], marked: set[str]) -> Iterable[OpaqueTask]:
        if mark in marked:
            return

        # Check that we actually have this task
        if mark not in self._tasks:
            cand = nearest_matching(mark, self.tasks, key=lambda t: t.name)
            raise InvalidTask(NoneSuch[OpaqueTask](mark, cand))

        # Append to the search stack
        stack.append(mark)
        t = self._tasks[mark]

        # Check if it was disabled:
        if t.is_disabled:
            assert t.disabled_reason
            raise DisabledTaskError(stack, t.disabled_reason)

        # Yield this item
        yield t
        marked.add(t.name)

        # Yield each hard dependency
        for d in t.depends:
            if d.is_order_only or d.dep_name in marked:
                continue
            yield from self._collect(d.dep_name, stack, marked)

        # Pop from the search stack
        stack.pop()

    def low_level_graph(self, marks: Iterable[str]) -> LowLevelDAG[Task[Any]]:
        by_name = {t.name: t for t in (itertools.chain.from_iterable(self._collect(m, [], set()) for m in marks))}
        return LowLevelDAG[Task[Any]](
            nodes=by_name.values(),
            edges=itertools.chain.from_iterable(
                ((by_name[d.dep_name], t) for d in t.depends if d.dep_name in by_name) for t in by_name.values()),
        )

    def create_executor(self,
                        marks: Iterable[str],
                        *,
                        exec_fn: Callable[[Task[T]], Awaitable[T]] = default_execute) -> exec.SimpleExecutor[Task[Any]]:
        graph = self.low_level_graph(marks)
        return TaskExecutor[Opaque](graph, exec_fn=cast(Any, exec_fn), catch_signals=True, fail_cancels=True)

    async def execute(
            self,
            marks: Iterable[str],
            *,
            exec_fn: Callable[[Task[T]], Awaitable[T]] = default_execute) -> Mapping[Task[Any], NodeResult[Task[Any]]]:
        ex = self.create_executor(marks, exec_fn=exec_fn)
        return await ex.run_all()


_CleanupFunc: TypeAlias = 'Callable[[], None|Awaitable[None]]'


class _ExecCtx(NamedTuple):
    exe: TaskExecutor[Any]
    task: Task[Any]
    add_graph_cleanup: Callable[[_CleanupFunc], None]
    add_cleanup_task: Callable[[asyncio.Task[Any]], None]
    cleanups: list[_CleanupFunc]


_TASK_CONTEXT = contextvars.ContextVar[_ExecCtx]('_TASK_CONTEXT')


def current_task() -> Task[Any]:
    return _TASK_CONTEXT.get().task


def cleanup(func: _CleanupFunc, *, when: Literal['now', 'graph-exit', 'task-exit'] = 'graph-exit') -> None:
    """
    Define a cleanup routine for the task or graph.

    :param func: A cleanup function to execute. If invoking this function
        returns an `~Awaitable` object, that object will be awaited before the
        task graph exits.
    :param when:
        Specify when the cleanup routine should begin executing. One of:

        - ``"graph-exit"`` - (Default) launch the cleanup routine only after no
          more tasks are running and the graph execution is shutting down.
        - ``"task-exit"`` - Launch the cleanup routine after the current task
          is finished executing.
        - ``"now"`` - Launch the cleanup routine immediately.

    Regardless of when a cleanup routine is launched, it will not block any
    process in graph execution except for graph shutdown, when all cleanup
    routines will be awaited.

    Use cleanup routines to execute actions that should be taken, but do not
    necessarily need to block the remainder of task execution. For example, the
    removal of temporary files should always occur, but it is not necessary that
    subsequent tasks need to wait for this to be complete.

    .. note::

        The cleanup function executes without a task context, so task-contextual
        actions must not be used!

    .. note::

        Exceptions that occur during the cleanup execution are ignored but do
        generate a `~RuntimeWarning`.

    .. note::

        Cleanup routines are executed asynchronously, but block graph execution
        shutdown. If a cleanup function takes too long to execute (Many
        seconds), it may be cancelled.
    """
    tctx = _TASK_CONTEXT.get()
    if when == 'task-exit':
        tctx.cleanups.append(func)
    elif when == 'graph-exit':
        tctx.add_graph_cleanup(func)
    elif when == 'now':
        t = asyncio.create_task(_run_cleanup(func), name=f'Cleanup {func!r} from task "{tctx.task.name}"')
        tctx.add_cleanup_task(t)
    else:
        _: int = when


async def _run_cleanup(cleaner: _CleanupFunc) -> None:
    r = cleaner()
    if inspect.isawaitable(r):
        await r


@overload
def result_of(task: Task[T]) -> Awaitable[T]:
    ...


@overload
def result_of(task: str) -> Awaitable[Any]:
    ...


def result_of(task: Task[Any] | str) -> Awaitable[Any]:
    """
    Obtain the result of the given task.

    The other task must be a direct dependency of the currently executing task,
    and must not be an order-only dependency.
    """
    ctx = _TASK_CONTEXT.get()
    ctx_task = ctx.task
    if isinstance(task, str):
        find = (n for n in ctx.exe.graph.all_nodes if n.name == task)
        try:
            task, = find
        except ValueError:
            raise NameError(f'No known task named "{task}"')
    find = (d for d in ctx_task.depends if d.dep_name == task.name)
    try:
        dep = first(find)
    except ValueError:
        raise NoDependencyError(ctx_task, task)
    if dep.is_order_only:
        raise IsOrderOnlyDependencyError(ctx_task, task)
    ares = ctx.exe.result_of(task)
    return _await_result(ares, task.name)


async def _await_result(ares: Awaitable[NodeResult[Any]], name: str) -> Any:
    res = await ares
    if isinstance(res.result, Success):
        return res.result.value
    raise RuntimeError(f'Attempted to obtain result of task "{name}", which is failed/cancelled')


class _GraphContext(NamedTuple):
    cleanups: list[_CleanupFunc]
    cleanup_tasks: list[asyncio.Task[Any]]


_GRAPH_CONTEXT = contextvars.ContextVar[_GraphContext]('_GRAPH_CONTEXT')


class TaskExecutor(exec.SimpleExecutor[Task[T]]):
    """
    A DAG executor that executes task objects and allows the results of tasks
    to be shared between each other using the `.result_of` function.
    """
    def __init__(self, graph: LowLevelDAG[Task[T]], exec_fn: Callable[[Task[T]], Awaitable[T]], *, catch_signals: bool,
                 fail_cancels: bool):
        super().__init__(graph, exec_fn)
        self.__graph = graph
        self.__catch_signals = catch_signals
        self.__fail_cancels = fail_cancels

    @property
    def graph(self):
        """The low-level graph executed by this executor"""
        return self.__graph

    @asynccontextmanager
    async def running_context(self) -> AsyncIterator[None]:
        from dagon.event.cancel import CancellationToken
        ctx = _GraphContext([], [])
        async with AsyncExitStack() as st:
            await st.enter_async_context(super().running_context())
            st.enter_context(scope_set_contextvar(_GRAPH_CONTEXT, ctx))
            tok = CancellationToken.get_context_local()
            if tok is None:
                tok = CancellationToken()
                tok.set_context_local(tok)
                st.callback(lambda: tok.set_context_local(None))
            handle_signals: list[int] = []
            if self.__catch_signals:
                handle_signals.extend((signal.SIGINT, signal.SIGTERM))
                if os.name == 'nt':
                    handle_signals.append(signal.SIGBREAK)
                else:
                    handle_signals.append(signal.SIGQUIT)
            st.enter_context(self._handle_signals(handle_signals, tok))
            yield

        cleanup_tasks = set(ctx.cleanup_tasks)
        # Start all graph-exit cleanups
        for clean in ctx.cleanups:
            cleanup_tasks.add(asyncio.create_task(_run_cleanup(clean)))
        # Collect cleanup results
        await self._cleanup(cleanup_tasks)

    async def _cleanup(self, cleanups: set[asyncio.Task[Any]]) -> None:
        # Newly added
        all_tasks: set[asyncio.Task[Any]] = set()
        if cleanups:
            all_tasks |= cleanups
            _done, pending = await asyncio.wait(cleanups, timeout=3, return_when=asyncio.ALL_COMPLETED)
            if pending:
                print('Cleanup tasks are executing...')
                _done, pending = await asyncio.wait(pending, timeout=12, return_when=asyncio.ALL_COMPLETED)

        for t in all_tasks:
            self._print_cleanup_info(t)

    @staticmethod
    def _print_cleanup_info(task: asyncio.Task[Any]) -> None:
        if task.done():
            try:
                task.result()
            except:
                tb = traceback.format_exc()
                warnings.warn(f'Cleanup task {task!r} exited with an exception:\n{tb}', RuntimeWarning)
            return
        stack = task.get_stack()
        if not stack:
            tb = '<no traceback could be generated>'
        else:
            tb = traceback.format_stack(stack[-1])
        warnings.warn(f'Cleanup {task!r} did not complete in a reasonable time: {tb}', RuntimeWarning)

    @contextmanager
    def _handle_signals(self, sigs: Sequence[int], cancel: CancellationToken) -> Iterator[None]:
        loop = asyncio.get_event_loop()
        prev_handlers = [signal.signal(n, lambda n, fr: self._do_stop_threadsafe(n, fr, cancel, loop)) for n in sigs]
        try:
            yield
        finally:
            for sig, prev in zip(sigs, prev_handlers):
                signal.signal(sig, prev)

    @staticmethod
    def _do_stop_threadsafe(_signum: int, _frame: None | FrameType, cancel: CancellationToken,
                            loop: asyncio.AbstractEventLoop) -> None:
        loop.call_soon_threadsafe(cancel.cancel)

    async def do_run_node(self, node: Task[T]) -> NodeResult[Task[T]]:
        """
        Overridden from `SimpleExecutor.do_run_node`. Sets up the context to
        allow result sharing between tasks. Calls `.do_run_task` to execute
        the task within the scope that allows for result sharing.
        """
        from ..event import CancellationToken
        ctx = _ExecCtx(self,
                       task=node,
                       add_graph_cleanup=_GRAPH_CONTEXT.get().cleanups.append,
                       add_cleanup_task=_GRAPH_CONTEXT.get().cleanup_tasks.append,
                       cleanups=[])
        with scope_set_contextvar(_TASK_CONTEXT, ctx):
            result = await self.do_run_task(node)
            if isinstance(result.result, Failure) and self.__fail_cancels:
                cancel = CancellationToken.get_context_local()
                if cancel:
                    cancel.cancel()
            for clean in ctx.cleanups:
                ctx.add_cleanup_task(asyncio.create_task(_run_cleanup(clean)))
        return result

    async def do_run_task(self, node: Task[T]) -> NodeResult[Task[T]]:
        return await super().do_run_node(node)


_CURRENT_DAG = contextvars.ContextVar['TaskDAG | None']('_CURRENT_DAG', default=None)


def current_dag() -> TaskDAG:
    """
    Get the current thread/task-local Dag instance.
    """
    dag = _CURRENT_DAG.get()
    if dag is None:
        raise RuntimeError('There is no dag set for the current thread')
    return dag


def set_current_dag(dag: TaskDAG | None) -> None:
    """
    Set the current thread-local Dag instance.
    """
    _CURRENT_DAG.set(dag)


@contextmanager
def populate_dag_context(dag: TaskDAG) -> Iterator[None]:
    """
    Set the current thread-local Dag to ``dag``, and then restore the old
    thread-local Dag value when the context manager exists.

    :rtype: ContextManager[None]
    """
    tok = _CURRENT_DAG.set(dag)
    try:
        yield
    finally:
        _CURRENT_DAG.reset(tok)



"""
Module for creating/managing persistent data.
"""

from __future__ import annotations

import contextvars
import json
from contextlib import asynccontextmanager, contextmanager
from typing import (TYPE_CHECKING, Any, AsyncIterator, Awaitable, Callable, Generic, Iterator, NamedTuple, Optional,
                    Sequence, overload)

from dagon import util
from dagon.ext.iface import OpaqueTaskGraphView
from dagon.task.dag import OpaqueTask
from dagon.util import T, U, typecheck, unused
from typing_extensions import Protocol

from .. import db
from ..ext.base import BaseExtension
from ..cache import Key

_DEFAULT_SENTINEL: Any = object()


class IPersistentDataAccess(Protocol):
    def set(self, key: Key[T], value: T) -> Awaitable[None]:
        ...

    @overload
    def get(self, key: Key[T]) -> Awaitable[T]:
        ...

    @overload
    def get(self, key: Key[T], *, default: U) -> Awaitable[T | U]:
        ...

    @overload
    def get(self, key: Key[T], *, default_factory: Callable[[], U | Awaitable[U]]) -> Awaitable[T | U]:
        ...

    def has(self, key: Key[Any]) -> Awaitable[bool]:
        ...


class _DatabasePersistor:
    def __init__(self, db_: db.Database, task_id: Optional[db.TaskID]) -> None:
        self._db = db_
        self._task = task_id

    def get(self,
            key: Key[T],
            *,
            default: T = _DEFAULT_SENTINEL,
            default_factory: Callable[[], T] = _DEFAULT_SENTINEL) -> Awaitable[T]:
        """
        Load a persisted value from the database.

        :param key: The key of the stored value.
        :param task_id: The task that they value is attached to. If ``None``,
            it is considered a "global" persisted value.

        :return: The persisted value, or the default if one is provided
        :raises: KeyError if no data was persisted with the given key and no
            default was provided to this call.
        """
        return self._get(key, default, default_factory)

    async def _get(self, key: Key[T], default: T, default_factory: Callable[[], T | Awaitable[T]]) -> T:
        data = self._db(
            r'''
            SELECT data
              FROM dagon_persists
             WHERE key IS :key
                   and task_id IS :task_id
            ''',
            key=key.key,
            task_id=self._task,
        )
        for dat_str, in data:
            return json.loads(dat_str)
        if default is not _DEFAULT_SENTINEL:
            return default
        if default_factory is not _DEFAULT_SENTINEL:
            return await util.ensure_awaitable(default_factory())
        raise KeyError(f'No value for "{key.key}" is persisted, and no default was specified')

    async def set(self, key: Key[T], value: T) -> None:
        """
        Store a persistent value in the database.

        :param key: The key under which to store the value.
        :param task_id: The task to attach the value to. If ``None``, the
            value is "global" and not attached to a task.
        :param value: The value to store. Must be JSON-serializable.
        """
        async with self._db.transaction_context():
            if await self.has(key):
                self._db('''
                        UPDATE dagon_persists
                           SET data = :data
                         WHERE key IS :key
                               AND task_id IS :tid
                        ''',
                         data=json.dumps(value),
                         key=key.key,
                         tid=self._task)
            else:
                self._db('''
                         INSERT INTO dagon_persists (key, task_id, data)
                         VALUES (:key, :tid, :data)
                         ''',
                         key=key.key,
                         tid=self._task,
                         data=json.dumps(value))

    def has(self, key: Key[Any]) -> util.ReadyAwaitable[bool]:
        return util.ReadyAwaitable(0 != util.cell(
            self._db('SELECT count(*) FROM dagon_persists WHERE key IS :key AND task_id IS :tid',
                     key=key.key,
                     tid=self._task)))


_TASK_LOCAL_PERSIST = contextvars.ContextVar[IPersistentDataAccess]('_TASK_LOCAL_PERSIST')
_GLOBAL_PERSIST = contextvars.ContextVar[IPersistentDataAccess]('_GLOBAL_PERSIST')


@contextmanager
def set_task_local_persistor(p: IPersistentDataAccess) -> Iterator[IPersistentDataAccess]:
    t = _TASK_LOCAL_PERSIST.set(p)
    try:
        yield p
    finally:
        _TASK_LOCAL_PERSIST.reset(t)


@contextmanager
def set_global_persistor(p: IPersistentDataAccess) -> Iterator[IPersistentDataAccess]:
    t = _GLOBAL_PERSIST.set(p)
    try:
        yield p
    finally:
        _GLOBAL_PERSIST.reset(t)


def get_task_local_persistor() -> IPersistentDataAccess:
    return _TASK_LOCAL_PERSIST.get()


def get_global_persistor() -> IPersistentDataAccess:
    return _GLOBAL_PERSIST.get()


class _ImplicitPersistenceItem(NamedTuple):
    cv_: contextvars.ContextVar[IPersistentDataAccess]

    def get(self, key: Key[Any], *, default: Any = _DEFAULT_SENTINEL, default_factory: Any = _DEFAULT_SENTINEL) -> Any:
        if default is not _DEFAULT_SENTINEL:
            return self.cv_.get().get(key, default=default)
        if default_factory is not _DEFAULT_SENTINEL:
            return self.cv_.get().get(key, default_factory=default_factory)
        return self.cv_.get().get(key)

    async def set(self, key: Key[T], value: T) -> None:
        await self.cv_.get().set(key, value)

    def has(self, key: Key[Any]) -> Awaitable[bool]:
        return self.cv_.get().has(key)


if TYPE_CHECKING:
    typecheck(IPersistentDataAccess)(_DatabasePersistor)
    typecheck(IPersistentDataAccess)(_ImplicitPersistenceItem)

_MIGRATIONS: Sequence[str] = [
    r'''
    CREATE TABLE dagon_persists (
        task_id REFERENCES dagon_tasks DEFAULT NULL,
        key TEXT NOT NULL,
        data TEXT NOT NULL
    )
    ''',
]


class _PersistExt(BaseExtension[None, None, None]):
    dagon_ext_name = 'dagon.persist'
    dagon_ext_requires = ['dagon.db']

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[None]:
        gdb = db.global_context_data()
        assert gdb
        with set_global_persistor(in_database(gdb.database)):
            yield

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[None]:
        gdb = db.global_context_data()
        tdb = db.task_context_data()
        assert tdb and gdb
        with set_task_local_persistor(in_database(gdb.database, task=tdb.task_id)):
            yield


unused(_PersistExt)

globl: IPersistentDataAccess = util.create_lazy_lookup(_GLOBAL_PERSIST.get)
local: IPersistentDataAccess = util.create_lazy_lookup(_TASK_LOCAL_PERSIST.get)


def in_database(db_: db.Database, *, task: db.TaskID | None = None) -> IPersistentDataAccess:
    db.apply_migrations(db_.sqlite3_db, 'dagon_persist_meta', _MIGRATIONS)
    return _DatabasePersistor(db_, task)



from datetime import datetime, timedelta
import json
from pathlib import Path

import pytest

from dagon import task

from .. import util
from .. import http
from tests.test_util import dag_test

JSON_PLACEHOLDER = 'https://jsonplaceholder.typicode.com'

requires_http = pytest.mark.skipif(not http.is_supported(), reason='Requires [http] extra')


@pytest.mark.asyncio
@requires_http
async def test_chunked_get() -> None:
    buf_iter = http.get_chunked(f'{JSON_PLACEHOLDER}/posts')
    buf_acc = b''
    async for buf in buf_iter:
        buf_acc += buf

    data = json.loads(buf_acc)
    assert isinstance(data, list)


@pytest.mark.asyncio
@requires_http
async def test_download() -> None:
    local_file = Path(__file__).parent / 'test-download.json'
    try:
        await http.download(f'{JSON_PLACEHOLDER}/posts', destination=local_file)
        data = json.loads(local_file.read_text())
        assert isinstance(data, list)
    finally:
        if local_file.exists():
            local_file.unlink()


@pytest.mark.asyncio
@requires_http
async def test_download_tmp() -> None:
    async with http.download_tmp(f'{JSON_PLACEHOLDER}/posts/1') as tmpfile:
        json_str = tmpfile.read_text()
        print(json_str)
        data = json.loads(json_str)
        assert isinstance(data, dict)


@dag_test()
@requires_http
def test_cached_downloads():
    url = 'https://github.com/Kitware/CMake/releases/download/v3.25.0-rc4/cmake-3.25.0-rc4-linux-x86_64.tar.gz'

    dur = timedelta()

    @task.define()
    async def download_file():
        nonlocal dur
        start = datetime.now()
        async with http.download_tmp(url):
            pass
        dur = datetime.now() - start

    @task.define(depends=[download_file])
    async def download_again():
        start = datetime.now()
        async with http.download_tmp(url):
            pass
        new_dur = datetime.now() - start
        assert new_dur < dur + timedelta(seconds=2)

    return [download_again]


@dag_test()
@requires_http
def test_download_task(tmp_path: Path):
    url = 'https://github.com/Kitware/CMake/releases/download/v3.25.0-rc4/cmake-3.25.0-rc4-linux-x86_64.tar.gz'

    get_cmake_str = http.download_task('get-str', url)
    get_cmake_fac = http.download_task('get-fac', lambda: url)
    get_cmake_afac = http.download_task('get-afac', lambda: util.ReadyAwaitable(url))
    get_cmake_newname = http.download_task('get-with-filename', url, filename='cmake.tgz')
    get_cmake_newdest = http.download_task('get-with-dest', url, destination=tmp_path / 'cmake.tar.gz')

    @task.define(depends=[get_cmake_str, get_cmake_fac, get_cmake_afac, get_cmake_newname, get_cmake_newdest])
    async def check():
        assert (await task.result_of(get_cmake_str)).name == 'cmake-3.25.0-rc4-linux-x86_64.tar.gz'
        assert (await task.result_of(get_cmake_fac)).name == 'cmake-3.25.0-rc4-linux-x86_64.tar.gz'
        assert (await task.result_of(get_cmake_afac)).name == 'cmake-3.25.0-rc4-linux-x86_64.tar.gz'
        assert (await task.result_of(get_cmake_newname)).name == 'cmake.tgz'
        assert (await task.result_of(get_cmake_newdest)).name == 'cmake.tar.gz'

    return [check]



from __future__ import annotations

import json
from contextlib import asynccontextmanager
from typing import (TYPE_CHECKING, Any, AsyncIterator, Generic, NamedTuple, NewType, TypeVar, cast, overload)

from typing_extensions import Final, Literal, Protocol, Self

from dagon import util

from .. import db
from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..task.dag import OpaqueTask
from ..util import JSONValue, T, U, create_lazy_lookup, typecheck

JSONValueT = TypeVar('JSONValueT', bound=JSONValue)

_DEFAULT_SENTINEL: Any = object()
_db_mod = db


class Key(Generic[T]):
    def __init__(self, key: str) -> None:
        self.key = key


class _GlobalContext(NamedTuple):
    db: _db_mod.Database
    tmp_db: _db_mod.Database
    run_cache: _DbCache
    cache: _DbCache


class _TaskContext(NamedTuple):
    run_cache: _DbCache
    cache: _DbCache


_CacheNodeID = NewType('_CacheNodeID', int)
_CacheDirID = NewType('_CacheDirID', int)

_MIGRATIONS: list[str] = [
    r'''
    CREATE TABLE dagon_cache_nodes (
        node_id INTEGER PRIMARY KEY,
        name TEXT NOT NULL,
        parent_id INTEGER REFERENCES dagon_cache_nodes(node_id)
            ON DELETE CASCADE,
        CONSTRAINT dagon_cache_node_uniqueness UNIQUE (name, parent_id)
    )
    ''',
    r'''
    CREATE TABLE dagon_cache_entries (
        entry_id INTEGER PRIMARY KEY,
        node_id INTEGER
            NOT NULL
            REFERENCES dagon_cache_nodes(node_id)
                ON DELETE CASCADE,
        key TEXT NOT NULL,
        value TEXT NOT NULL,
        CONSTRAINT dagon_cache_entry_uniqueness
            UNIQUE(key, node_id)
    )
    ''',
    r'''
    CREATE TABLE dagon_cache_dirs (
        dir_id INTEGER PRIMARY KEY,
        node_id INTEGER
            NOT NULL
            REFERENCES dagon_cache_nodes,
        parent_id INTEGER REFERENCES dagon_cache_dirs(dir_id),
        name TEXT NOT NULL,
        UNIQUE (parent_id, name)
    )
    ''',
    r'''
    CREATE TABLE dagon_cache_files (
        file_id INTEGER PRIMARY KEY,
        dir_id INTEGER
            NOT NULL
            REFERENCES dagon_cache_dirs(dir_id),
        name TEXT NOT NULL,
        UNIQUE (dir_id, name)
    )
    ''',
]


class _Ext(BaseExtension[None, _GlobalContext, _TaskContext]):
    dagon_ext_name = 'dagon.cache'
    dagon_ext_requires = ['dagon.db']

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[_GlobalContext]:
        gdb = db.global_context_data()
        assert gdb
        gdb = gdb.database
        persistent_id = _DbCache.prepare_db(gdb, None)
        tmpdb = db.Database.open('')
        tmp_id = _DbCache.prepare_db(tmpdb, None)
        yield _GlobalContext(
            gdb,  #
            tmpdb,
            _DbCache(gdb, _CacheNodeID(tmp_id)),
            _DbCache(gdb, _CacheNodeID(persistent_id)))

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[_TaskContext]:
        globl = self.global_data()
        gdb = globl.db
        tmpdb = globl.tmp_db
        persistent_id = _DbCache.prepare_db(gdb, task.name)
        tmp_id = _DbCache.prepare_db(tmpdb, task.name)
        yield _TaskContext(
            _DbCache(tmpdb, _CacheNodeID(tmp_id)),  #
            _DbCache(gdb, _CacheNodeID(persistent_id)))


class ICache(Protocol):
    @overload
    def child(self, key: str) -> Self:
        ...

    @overload
    def child(self, key: str, *, if_missing: Literal['create', 'fail']) -> Self:
        ...

    @overload
    def child(self, key: str, *, if_missing: Literal['ignore']) -> Self | None:
        ...

    def remove_child(self, key: str, *, absent_ok: bool = False) -> None:
        ...

    def set(self, key: Key[JSONValueT], value: JSONValueT) -> JSONValueT:
        ...

    @overload
    def pop(self, key: Key[JSONValueT]) -> JSONValueT:
        ...

    @overload
    def pop(self, key: Key[JSONValueT], *, default: U) -> JSONValueT | U:
        ...

    @overload
    def get(self, key: Key[JSONValueT]) -> JSONValueT:
        ...

    @overload
    def get(self, key: Key[JSONValueT], *, default: U) -> JSONValueT | U:
        ...


class _DbCache(ICache):
    def __init__(self, db_: db.Database, node_id: _CacheNodeID) -> None:
        self._node = node_id
        self._db = db_

    @staticmethod
    def prepare_db(db_: db.Database, task_: str | None) -> int:
        db.apply_migrations(db_.sqlite3_db, 'dagon_cache_meta', _MIGRATIONS)
        # Create the root for graphs, and a root for tasks
        db_(r'''
            INSERT OR IGNORE INTO dagon_cache_nodes (node_id, name)
            VALUES (0, '@graph'), (1, '@task')
            ''')
        if task_ is None:
            return 0
        db_(
            r'''
            INSERT OR IGNORE INTO dagon_cache_nodes (name, parent_id)
            VALUES (:task, 1)
            ''',
            task=task_,
        )
        task_id = util.cell(
            db_(
                r'''
                SELECT node_id FROM dagon_cache_nodes
                WHERE name=:task AND parent_id=1
                ''',
                task=task_,
            ))
        return task_id

    def child(self, key: str, *, if_missing: Literal['create', 'fail', 'ignore'] = 'fail') -> _DbCache | None:
        if if_missing == 'create':
            self._db(
                r'''
                INSERT OR IGNORE INTO dagon_cache_nodes
                    (name, parent_id)
                VALUES (:name, :parent)
                ''',
                name=key,
                parent=self._node,
            )
        child_id: int | None = util.cell(
            self._db(
                r'''
                SELECT node_id
                  FROM dagon_cache_nodes
                 WHERE name=:name AND parent_id=:parent
                ''',
                name=key,
                parent=self._node,
            ),
            default=None,
        )
        if child_id is None:
            if if_missing == 'fail':
                raise LookupError(key)
            return None
        return _DbCache(self._db, _CacheNodeID(child_id))

    def remove_child(self, key: str, *, absent_ok: bool = False) -> None:
        before = self._db.sqlite3_db.total_changes
        self._db(
            r'''
            DELETE FROM dagon_cache_nodes
             WHERE name=:name AND parent_id=:parent
            ''',
            name=f'.{key}',
            parent=self._node,
        )
        after = self._db.sqlite3_db.total_changes
        if before == after and not absent_ok:
            raise LookupError(key)

    def set(self, key: Key[JSONValueT], value: JSONValueT) -> JSONValueT:
        encoded = json.dumps(value, sort_keys=True)
        self._db(
            r'''
            INSERT OR REPLACE INTO dagon_cache_entries
                (node_id, key, value)
            VALUES (:node, :key, :val)
            ''',
            node=self._node,
            key=key.key,
            val=encoded,
        )
        return value

    def _get(self, key: str) -> None | tuple[int, str]:
        return util.first(
            items=self._db(
                r'''
                SELECT entry_id, value FROM dagon_cache_entries
                WHERE node_id=:node AND key=:key
                ''',
                node=self._node,
                key=key,
            ),
            default=None,
        )

    def get(self, key: Key[JSONValueT], *, default: JSONValueT = _DEFAULT_SENTINEL) -> JSONValueT:
        pair = self._get(key.key)
        if pair is None:
            if default is _DEFAULT_SENTINEL:
                raise KeyError(key.key)
            return default
        v = json.loads(str(pair[1]))
        return v

    def pop(self, key: Key[JSONValueT], *, default: JSONValueT = _DEFAULT_SENTINEL) -> JSONValueT:
        pair = self._get(key.key)
        if pair is None:
            if default is _DEFAULT_SENTINEL:
                raise KeyError(key.key)
            return default
        self._db('DELETE FROM dagon_cache_entries WHERE entry_id=:id', id=pair[0])
        v = json.loads(str(pair[1]))
        return v


if TYPE_CHECKING:
    typecheck(ICache)(_DbCache)

task_run: Final[ICache] = create_lazy_lookup(lambda: _Ext.task_data().run_cache)
graph_run: Final[ICache] = create_lazy_lookup(lambda: _Ext.global_data().run_cache)
task: Final[ICache] = create_lazy_lookup(lambda: _Ext.task_data().cache)
graph: Final[ICache] = create_lazy_lookup(lambda: _Ext.global_data().cache)
